#!/bin/bash
# ITRS OP5 Monitor installer.
# Copyright (C) 2021  OP5 AB
""":"
python=$(type -p python3 2>/dev/null)
[ $? -eq 0 ] || python=/usr/libexec/platform-python
if ! [ -x $python ]; then
    echo "python3 is required to continue installation"
    dnf install python3
    python=$(type -p python3)
fi
[ -x "$python" ] && exec $python $0 "$@"
echo "error: please install python3 and re-run this script"
exit 126
"""

from __future__ import print_function
import platform
import sys

MIN_PYTHON_VERSION = (3, 6)
if sys.version_info < MIN_PYTHON_VERSION:
    print(
        "Installed version of {} ({}) is too low, need {} to continue".format(
            sys.executable,
            platform.python_version(),
            ".".join(map(str, MIN_PYTHON_VERSION)),
        ),
        file=sys.stderr,
    )
    sys.exit(1)

import argparse
import collections
import configparser
import enum
import functools
import json
import logging
import os
import pathlib
import re
import shlex
import shutil
import subprocess
import textwrap
import time

from typing import Optional

PRODUCT = "op5-monitor"  # i.e. "ITRS OP5 Monitor"
VERSION = "9rolling"
SUPPORT_NAME = "ITRS Group support"
USAGE_TEXT = f"""\
Run %(prog)s without arguments to install {PRODUCT} in interactive mode.
Use command line options listed above to alter the behaviour.

By default {PRODUCT} is installed from PRM packages that are bundled in
the same archive that this installer script.

"""
SUPPORTED_OS = {
    "centos": {"versions": [8]},
    "rocky": {"versions": [8]},
    "rhel": {"versions": [8]},
}
SCRIPT_DIR = pathlib.Path(__file__).parent.resolve()
DEFAULT_RPM_DIR = SCRIPT_DIR / "rpm"
config = None
dnf = None
DNF_OPTIONS = os.environ.get("DNF_OPTIONS")
TMPDIR = os.environ.get("OP5_TMPDIR", "/var/tmp")
_SENTINEL_REPO = "_SENTINEL_REPO"
OP5_RPM_GPGKEY_ID = "gpg-pubkey-f6e3ade7-46f67df7"
OP5_RPM_GPGKEY = """\
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1.4.5 (GNU/Linux)

mQGiBEb2ffcRBACdwwaKKRkPvqoCMpB7qefP2ga/O+bUgrN6Jl4Ml2fzYoUFzI7d
0RsQF7yOnYSobiS09gSpkD5J3TyFPkAGEpntVw3diXZMdwHmWlUftegRQlIBVuMl
a3m7NnxOeyKVt8dZNLJQmapvbh8Qr8Bb9H0wQOi+8EFBMr/uSfr4rXwsGwCglSeg
BelnEYGVUINBiTnp4aIXsbsD/3GFTkkV4s0TC+II7VfmB7N7+dJ587pAMgxa4oJ8
RRhjXlgBOWtHgwKR/q7qE8By0XrQ4jUqHvWWRDUZTWE0H0MXsApkv92qptBCxWAV
BBfQTnpjVqPwbo75HFe93eHCVlluGLJO2DGiJuGUje8Vipy0LRZAHQq0rZYC/jwW
5VG+A/kBJUxQkdvkKlQzwKaFByvCBcj20KySsoV/d/K8oPUiSO/Irdt8w8JCaBI2
t0gCOvVu71RiiyLtHXjYmsL5oyC5B7CUGHONeUeu+RB09W8uG/h5qXl97Pdln2JP
01wFhumHoc699BgNUC8Td3gPdyuvUm/qtNjL+BdqM97wczKfyLQtb3A1IEFCIChv
cDUgQUIgcnBtIHNpZ25pbmcpIDxzdXBwb3J0QG9wNS5jb20+iGAEExECACAFAkb2
ffcCGwMGCwkIBwMCBBUCCAMEFgIDAQIeAQIXgAAKCRAvJHfz9uOt56B7AJ461ub4
qxCtcJ8wEkTWsWiaco1/jgCfTAYgqefOgqVZ060Bcv4DXDw9j5Q=
=zQcj
-----END PGP PUBLIC KEY BLOCK-----
"""
DNF_VAR_OP5 = "DNF_VAR_GPGKEY_OP5"

log = logging.getLogger("main")
dlog = logging.getLogger("debug")


class _Config:

    _default_loglevel = getattr(
        logging,
        os.environ.get("OP5_LOGLEVEL", "WARNING").upper(),
        logging.WARNING,
    )
    _colors = {
        "red": "\033[31m",
        "green": "\033[32m",
        "yellow": "\033[33m",
        "blue": "\033[34m",
        "magenta": "\033[35m",
        "cyan": "\033[36m",
        "white": "\033[37m",
        "reset": "\033[0m",
    }

    def __init__(
        self,
        *,
        interactive: bool,
        verbose: int,
        debug: bool,
        force: bool,
        assumeyes: bool,
        custom_repo,
        dnf_config: str,
        color_output: bool,
        repo_dir: Optional[pathlib.Path],
    ):
        self.interactive = interactive
        self.debug = debug
        self.force = force
        self.assumeyes = not interactive or assumeyes
        self.custom_repo = custom_repo
        self.dnf_config = dnf_config
        self.color_output = color_output and sys.stdout.isatty()
        self.custom_repo_dir = repo_dir
        if repo_dir:
            self.repo_dir = repo_dir.resolve()
        else:
            self.repo_dir = DEFAULT_RPM_DIR / "monitor/el/$releasever/$basearch"

        # Set log level to debug if debug is used, verbosity count,
        # OP5_LOGLEVEL or a default.
        if debug:
            self.log_level = logging.DEBUG
        elif verbose:
            self.log_level = max(self._default_loglevel - verbose * 10, logging.DEBUG)
        else:
            self.log_level = self._default_loglevel

        self.workdir = pathlib.Path(TMPDIR) / "op5-{}".format(
            time.strftime("%Y%m%d%H%M%S")
        )
        self.workdir.mkdir(0o700)
        self.debug_log_filename = self.workdir / "debug.log"
        self.debug_log_file = self.debug_log_filename.open("w")

        self.print_console = sys.stdout.isatty()

        self.os_info: Optional[OSInfo] = None

    def color(self, color):
        if self.color_output:
            return self._colors.get(color, self._colors["reset"])
        return ""

    def __del__(self):
        try:
            self.debug_log_file.close()
        except Exception:
            pass

    def log_all(self, logger=log):
        logger.debug("Active config settings")
        for k, v in self.__dict__.items():
            logger.debug("\t%s = %r", k, v)


class OP5Error(Exception):
    """Installer errors.

    Args:
        message: error message.
        *args: arguments for %s formatting of message.
        retcode: attach a return code to exit the program with.
    """

    _default_retcode = 1

    def __init__(self, message, *args, retcode=None):
        self._message = message
        self._args = args
        self.retcode = retcode if retcode is not None else self._default_retcode
        super().__init__(message, *args)

    def __str__(self):
        return self._message % self._args


class OSInfo:
    def __init__(self, name, id_, version):
        self.name = name
        self.id = id_
        self.version = version
        self.version_str = ".".join(map(str, self.version))

    def __str__(self):
        return f"{self.name} {self.version_str} ({self.id})"

    @classmethod
    def from_file(cls, file="/etc/os-release"):
        fields = {k: None for k in {"NAME", "ID", "VERSION_ID"}}
        dlog.info("Reading %s", file)
        try:
            with open(file) as fp:
                for line in fp:
                    dlog.debug("\tRead line:  %s", line.strip())
                    for key in fields:
                        m = re.match(r"""%s=(["']?)(.+)\1""" % key, line)
                        if m:
                            fields[key] = m[2]
        except OSError as exc:
            raise OP5Error("Failed to read os-release: %s", exc) from exc
        dlog.info("Read fields: %s", fields)
        missing = {k for k, v in fields.items() if v is None}
        if missing:
            raise OP5Error("Missing fields from %s: %s", file, ", ".join(missing))

        return cls(
            fields["NAME"], fields["ID"], make_version_tuple(fields["VERSION_ID"])
        )


class DNF:

    _REPOINFO_FIELDS = ("Repo-id", "Repo-name", "Repo-status", "Repo-filename")

    def __init__(self, config_):
        self._binary = "/usr/bin/dnf"
        self._flags = []
        self._config = config_

        # Use --dnf-config if set, otherwise rely of default.
        if self._config.dnf_config:
            self._conf = pathlib.Path(self._config.dnf_config).resolve(strict=True)
            self._flags.append(f"--config={self._conf}")

        if self._config.assumeyes:
            self._flags.append("--assumeyes")
        if self._config.debug:
            self._flags.append("-v")
            self._flags.append("--rpmverbosity=debug")
        if not self._config.color_output:
            self._flags.append("--color=never")
        if DNF_OPTIONS:
            self._flags.append(DNF_OPTIONS)

    def mk_cmd(self, *args):
        return "{} {} {}".format(
            self._binary,
            " ".join(self._flags),
            " ".join(args),
        )

    def repoinfo(self):
        """Return dict with configured DNF repos."""
        log.debug("Get repo info")
        proc = command_output(self.mk_cmd("repoinfo --all"), check=True)
        repos = collections.defaultdict(dict)
        repoid = None
        for line in proc.stdout.splitlines():
            m = re.match(r"\s*(\w\S+)\s*:\s*(.+?)$", line)
            if not m:
                repoid = None
                continue
            if m[1] == "Repo-id":
                repoid = m[2]
            if repoid and m[1] in self._REPOINFO_FIELDS:
                repos[repoid][m[1]] = m[2]
        dlog.debug("Configured DNF repos: %s", repos)
        return repos

    def setup_3rd_party_repos(self):
        log.info("Setup DNF repo: powertools")
        installed_repos = self.repoinfo()
        # In early CentOS 8 releases the repo id was "PowerTools", so find it
        # whatever the id is.
        for repoid, settings in installed_repos.items():
            if repoid.lower() == "powertools":
                break
            elif repoid.lower() == "codeready-builder-for-rhel-8-x86_64-rpms":
                break
        else:
            raise OP5Error("Cannot find DNF repository: powertools")
        command(self.mk_cmd("config-manager --dump %s" % repoid), check=True)
        self._enable_repo(settings)
        # For RHEL install the RPM directly from the URL
        if config.os_info.id == "rhel":
            epel_release_url = (
                "https://dl.fedoraproject.org/pub/epel/epel-release-latest-8.noarch.rpm"
            )
            command(self.mk_cmd(f"install {epel_release_url}"), stdout=True, check=True)
            return

        # For other distros we can install the needed 3rd party repos from
        # packages included in extras
        repos_to_install = (("epel", "epel-release"),)
        for repoid, repopkg in repos_to_install:
            log.info("Setup rpm repo: %s", repoid)
            if repoid not in installed_repos:
                info(f"Installing repo {repoid} ...")
                info("\n***** External command output")
                command(self.mk_cmd(f"install {repopkg}"), stdout=True, check=True)
                info("\n***** End external command\n")
            installed_repos = self.repoinfo()
            self._enable_repo(installed_repos[repoid])

    def _enable_repo(self, repo):
        dlog.debug("trace: enable repo: %s", repo)
        if repo["Repo-status"] == "enabled":
            log.info("%s is already enabled", repo["Repo-id"])
        else:
            info(
                "Enabling repository '%(Repo-name)s' (%(Repo-id)s) "
                "from '%(Repo-filename)s' ..." % repo
            )
            command(
                self.mk_cmd("config-manager --set-enabled %s" % repo["Repo-id"]),
                check=True,
            )


class Answer(enum.Enum):
    NONE = 1
    YES = 2
    NO = 3


def yes_no(message: str, selected=Answer.NONE, default=Answer.YES) -> Answer:
    if config.interactive and not config.force:
        assert sys.stdin.isatty()
        if selected == Answer.NONE:
            pre_selected = ""
        else:
            pre_selected = f"({selected.name.lower()})"

        prompt = f"\n{message} [y]es/[n]o {pre_selected}: "
        for _ in range(3):
            resp = input(prompt).lower()
            if resp in ("yes", "y"):
                return Answer.YES
            elif resp in ("no", "n"):
                return Answer.NO
            elif resp == "" and selected != Answer.NONE:
                # User hit enter to accept pre-selected default.
                return selected
            else:
                info(
                    "type one of yes/y/no/n{} to continue.".format(
                        " or hit enter to accept the default choice"
                        if selected != Answer.NONE
                        else ""
                    )
                )
        else:
            raise OP5Error("invalid response")
    elif config.force:
        info(f"""--force is in effect, would've prompted "{message}".""")
        return Answer.YES
    else:
        log.info(
            'Would prompt in interactive mode: "%s", use default answer: %s',
            message,
            default.name,
        )
        return default


def main():
    os.environ["LANG"] = "C.utf8"
    os.environ["LC_ALL"] = "C.utf8"
    parser = argparse.ArgumentParser(
        description="Installation program for {} {}".format(PRODUCT, VERSION),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=USAGE_TEXT,
    )
    parser.add_argument(
        "--noninteractive",
        dest="interactive",
        action="store_false",
        help=f"Install or update {PRODUCT} using default settings without prompting "
        "for user input. Implies --assumeyes.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase console logging verbosity, can be used twice (-vv). "
        "A debug log is always written to as well.",
    )
    parser.add_argument(
        "--assumeyes",
        action="store_true",
        help="Pass --assumeyes to DNF package manager",
    )
    parser.add_argument(
        "-C",
        "--no-color",
        action="store_false",
        dest="color_output",
        help="Disable ANSI colors in console output",
    )
    parser.add_argument(
        "--custom-repo",
        metavar="CONFIG",
        nargs="?",
        const=_SENTINEL_REPO,
        help="Don't use rpm repos bundled in installer archive. If set with a "
        "%(metavar)s value, add repos from that file. Without a value, assume that "
        "required repos are already configured on the system. Can be used to run an "
        "online-only installation from repo server.",
    )
    parser.add_argument(
        "--repo-dir",
        metavar="DIR",
        type=functools.partial(existing_path, is_dir=True),
        help="Custom path to RPM repository (dir containing 'repodata/')."
        " By default RPMs included in the installer archive are used.",
    )
    parser.add_argument(
        "--dnf-config",
        metavar="FILE",
        type=existing_path,
        help="Path to custom config file for package manager. It's also possible to "
        "customize DNF behaviour by passing CLI options in 'DNF_OPTIONS' environment "
        "variable.",
    )
    parser.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--force", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "-V", "--version", action="version", version=f"{PRODUCT} {VERSION} installer"
    )
    args = parser.parse_args()
    global config, dnf
    config = _Config(**vars(args))
    init_loggers(config)
    config.log_all()

    log.info(
        "Workdir: %s, debug logfile: %s", config.workdir, config.debug_log_filename
    )
    sanity_check()
    dlog.info(
        "Start %s on host %s (%s), %s version %s",
        sys.argv[0],
        platform.node(),
        platform.platform(),
        PRODUCT,
        VERSION,
    )
    dlog.debug("cli args: %s", args)
    dlog.debug("environment:\n%s", json.dumps(dict(os.environ.items()), indent=4))
    config.os_info = OSInfo.from_file()
    config.log_all(dlog)
    dnf = DNF(config)

    if config.interactive and not sys.stdin.isatty():
        raise OP5Error(
            "stdin is not connected to a tty. Run again in a terminal, "
            'or use "--noninteractive" option for an unattended installation, '
            "accepting defaults."
        )

    os.umask(0o0022)

    dlog.debug("Log all installed packages before installation begins")
    command("rpm --query --all", check=True)
    start_install()


def start_install():
    info("")
    info(
        "Welcome to the {} {} installation suite.".format(PRODUCT, VERSION),
        always_print=True,
    )
    info("\nThis script will guide you through the process.\n")

    update = command("rpm -q op5-monitor") == 0
    if update:
        info(f"{PRODUCT} is already installed - Update to {VERSION}")

    check_prerequisites()

    msg = textwrap.dedent(
        """
    This installer will install packages from {}, EPEL and OP5 repositories,
    and make necessary configuration adjustments on the system.
    The process can not be undone.
    """
    ).format(config.os_info.name)
    info(msg)
    if not config.assumeyes:
        info(
            textwrap.dedent(
                """\
            DNF will prompt you before doing certain operations. To skip being prompted,
            run the installer again with --noninteractive or --assumeyes options.
            """
            )
        )
    if (
        yes_no(
            "Do you want to continue with the installation?",
            default=Answer.YES,
        )
        == Answer.NO
    ):
        return False

    if not update and monitor_has_been_installed():
        for line in textwrap.wrap(
            "There are traces of {0} previously being installed on this system. "
            "It's recommended to install {0} on fresh system to ensure "
            "that all packages and their dependencies are installed as "
            "intended.".format(PRODUCT)
        ):
            log.warning(line)
        if (
            yes_no(
                "Are you sure you want to continue installation (not recommended)?",
                selected=Answer.NO,
                default=Answer.NO,
            )
            == Answer.NO
        ):
            return False

    if not update:
        configure_selinux(mode="permissive")
    configure_module_stream(update, "mariadb-server", "mariadb", "10.5")
    configure_module_stream(update, "php", "php", "7.4")
    check_mariadb_root()
    import_rpm_keys()
    dnf.setup_3rd_party_repos()
    install_op5_monitor(update)

    if update:
        info(
            "\nUpdate of {} to version {} completed successfully!\n".format(
                PRODUCT, VERSION
            ),
            always_print=True,
        )
        return
    info(
        textwrap.dedent(
            """
            Installation of {} {} completed successfully!

            It's recommended to reboot the system.
            """
        ).format(PRODUCT, VERSION),
        always_print=True,
    )


def install_op5_monitor(update):
    repos_dir = config.workdir / "repos.d"
    repos_dir.mkdir()
    opts = "--setopt=reposdir='/etc/yum.repos.d,%s'" % repos_dir
    if config.custom_repo == _SENTINEL_REPO:
        log.info("Assuming required repos are configured on the system")
        opts = ""
    elif config.custom_repo is not None:
        log.info("Copy repo config %s to %s", config.custom_repo, repos_dir)
        repo_file = shutil.copy(config.custom_repo, repos_dir / "custom.repo")
    else:
        repo_file = repos_dir / "op5-install.repo"
        log.debug("Write repo config to %s", repo_file)
        repo_file.write_text(
            textwrap.dedent(
                """\
                [op5-installer]
                name={product} installer
                baseurl=file://{cfg.repo_dir}
                gpgcheck=1
                gpgkey={gpgkey}
                """
            ).format(
                cfg=config,
                gpgkey=os.environ[DNF_VAR_OP5],
                product=f"{PRODUCT} {VERSION}",
            )
        )

    # Log repos in debug log
    command(
        dnf.mk_cmd("{} repolist --enabled -v".format(opts)),
    )

    info("\n***** Package installation start")
    command(dnf.mk_cmd(opts + " clean metadata"), stdout=True)
    if update:
        command(dnf.mk_cmd(opts + " update"), stdout=True, check=True)
    else:
        command(
            dnf.mk_cmd(opts + " install @op5-monitor op5-monitor-release"),
            stdout=True,
            check=True,
        )
    info("\n***** Package installation end\n")

    log.info("Check if firewalld is running ...")
    fw_enable_cmd = "firewall-cmd --permanent --add-service=op5-monitor"
    proc = command_output("firewall-cmd --state", stderr=subprocess.STDOUT)
    if proc.returncode == 0 and update:
        log.info(
            "Not configuring firewalld on updates. To manually permitting "
            'op5-monitor, run "%s" followed by "firewall-cmd --reload"',
            fw_enable_cmd,
        )
    elif proc.returncode == 0:
        log.info(
            "firewalld is running. Add op5-monitor service with command: %s",
            fw_enable_cmd,
        )
        if (
            yes_no(
                "Enable op5-monitor service in firewall config (recommended)?",
                selected=Answer.YES,
            )
            == Answer.YES
        ):
            command(fw_enable_cmd)
            command("firewall-cmd --reload")
        else:
            info(
                "Consider manually allowing op5-monitor in firewalld after "
                "installation."
            )
            info("run: '%s' followed by 'firewall-cmd --reload'" % fw_enable_cmd)
    else:
        log.warning(
            "firewalld service is NOT running. firewall-cmd --state: %s",
            proc.stdout.strip(),
        )
        log.warning(
            "It's recommended to enable firewalld and then run the following command "
            "to permit %s: '%s'",
            PRODUCT,
            fw_enable_cmd,
        )


def configure_module_stream(update: bool, rpm_name, stream_name, stream_version):
    """Active a non-default module stream on the system.

    Arguments:
        - update: whether this is an update on an existing system
        - rpm_name: name of a rpm installed from the module stream
        - stream_name: name of the module stream
        - stream_version: version of the module stream
    """
    log.info("Check if %s is already installed...", rpm_name)
    proc = command_output(f"rpm -q {rpm_name} --qf '%{{VERSION}}'")

    if proc.returncode != 0:
        log.info("%s is not installed - enable module stream", rpm_name)
        info(f"Enabling DNF module stream {stream_name}:{stream_version} ...")
        try:
            mod_conf = read_ini(
                pathlib.Path(f"/etc/dnf/modules.d/{stream_name}.module")
            )
        except FileNotFoundError as exc:
            dlog.debug("%s, no need to reset module", exc)
        else:
            if mod_conf[stream_name]["stream"]:
                log.debug("dnf module %s has stream enabled, reset module", stream_name)
                command(
                    dnf.mk_cmd(f"module reset {stream_name}"), stdout=True, check=True
                )
        command(
            dnf.mk_cmd(f"module enable {stream_name}:{stream_version}"),
            stdout=True,
            check=True,
        )
        return

    log.info("%s is already installed. Version: %s", rpm_name, proc.stdout)
    if not update and make_version_tuple(proc.stdout) < make_version_tuple(
        stream_version
    ):
        info(
            textwrap.dedent(
                """\
            %(rpm)s is already installed, with a version too old for %(product)s.
            We can try to upgrade the %(rpm)s installation, using a newer
            module stream, or you can abort the installation and manually
            uninstall %(rpm)s, and then re-run this installer.
            """
            )
            % {"rpm": rpm_name, "product": PRODUCT}
        )
        if (
            yes_no(
                "Attempt upgrading %s from module stream %s:%s ?"
                % (rpm_name, stream_name, stream_version),
                default=Answer.NO,
            )
            == Answer.NO
        ):
            raise OP5Error(
                "Cannot install %s while %s %s is installed",
                PRODUCT,
                rpm_name,
                proc.stdout,
            )
        # dnf module switch-to is not available on older CentOS 8 versions, but
        # this is really an edge case..
        command(
            dnf.mk_cmd(f"module switch-to {stream_name}:{stream_version}"),
            stdout=True,
            check=True,
            error_hint='"dnf module switch-to" command might not be available. '
            'Run "dnf update" and try again.',
        )


def check_mariadb_root():
    proc = command_output("rpm -q mariadb-server --qf '%{VERSION}'")

    # return if mariadb was not already installed
    if proc.returncode != 0:
        return

    log.info("Check authentication as root to existing MariaDB server...")
    if command("systemctl is-active mariadb") != 0:
        log.warning(
            "MariaDB server is installed, but not running, so we cannot verify "
            "that we can authenticate as root user to MariaDB."
        )
        return
    command(
        "mariadb --user=root --database=mysql -e status",
        check=True,
        error_hint="Failed to authenticate to existing MariaDB installation.",
    )


def configure_selinux(mode="permissive"):
    if command("/usr/sbin/selinuxenabled") == 0:
        info(f"Put SELinux in {mode} mode...", always_print=True)
        command("setenforce %s" % mode.capitalize(), check=True)
    else:
        log.info("SELinux is disabled")
    replace_line_in_file("/etc/selinux/config", r"^SELINUX=.*$", f"SELINUX={mode}")


def replace_line_in_file(
    filename, re_pattern, re_replacement, flags=re.MULTILINE, backup=True
) -> int:
    # Not for very large files, as contents are slurped into memory.
    # File must have the default encoding (utf-8).
    dlog.debug("trace: %r %r %r flags=%r", filename, re_pattern, re_replacement, flags)
    with open(filename) as fp:
        content = fp.read()
    dlog.debug("Old contents of %s:\n%s", filename, content[:5000])
    content, count = re.subn(re_pattern, re_replacement, content, flags=flags)
    if count == 0:
        log.info("No replacements made in %s", filename)
        return 0
    backup_filename = "%s.backup-%s" % (filename, time.strftime("%Y%m%d%H%M%S"))
    dlog.debug("copy '%s' -> '%s'", filename, backup_filename)
    shutil.copy2(filename, backup_filename)
    try:
        with open(filename, "w") as fp:
            fp.write(content)
    except OSError as exc:
        try:
            os.remove(filename)
        except OSError:
            pass
        os.rename(backup_filename, filename)
        dlog.exception("failed to write to %s", filename)
        raise OP5Error("Failed to edit file %s: %s", filename, exc) from exc
    if backup:
        log.info("Saved old %s to %s", filename, backup_filename)
    else:
        try:
            os.remove(backup_filename)
        except OSError as exc:
            log.warning("%s", exc)

    return count


def read_ini(path: pathlib.Path) -> dict:
    dlog.debug("trace: read_ini path=%s", path)
    text = path.read_text()
    dlog.debug("%s data: %s", path, text)
    cp = configparser.ConfigParser()
    cp.read_string(text, source=str(path))
    result = {}
    for s in cp.sections():
        result[s] = dict(cp.items(s))
    dlog.debug("processed data: %s", result)
    return result


def make_version_tuple(version: str) -> tuple:
    match = re.match(r"\d[\d.]*", version)
    if not match:
        raise OP5Error("Invalid version string: %s", version)
    version = match[0].rstrip(".")
    return tuple(int(n) for n in version.split("."))


def check_prerequisites():
    info("Checking prerequisites...")

    log.info("Check that we're running on a supported operating system...")
    try:
        supported_os = SUPPORTED_OS[config.os_info.id]
    except KeyError:
        raise OP5Error("Unsupported operating system: %s", config.os_info)
    if config.os_info.version[0] not in supported_os["versions"]:
        raise OP5Error(
            "Unsupported %s version: %s",
            config.os_info.name,
            config.os_info.version_str,
        )
    log.info("OS: %s", config.os_info)

    command(dnf.mk_cmd("--version"), check=True)
    command(dnf.mk_cmd("config-manager --dump"), check=True)

    if config.custom_repo:
        return

    log.info("Check package repository %s", config.repo_dir)
    proc = command_output(
        dnf.mk_cmd(
            "'--repofrompath=op5temp,{}' --repo op5temp makecache".format(
                config.repo_dir
            )
        )
    )
    if proc.returncode != 0:
        log.error(
            'Failed to make metadata cache of DNF repository "%s".', config.repo_dir
        )
        log.error("DNF error:")
        for line in proc.stderr.splitlines():
            log.error("  stderr: %s", line)
        if config.custom_repo_dir:
            log.error(
                'Selected repository "%s" is not a valid DNF repository',
                config.repo_dir,
            )
        if not config.force:
            raise OP5Error("Repository check failed")


def import_rpm_key_file(path, friendly_name):
    if (
        yes_no(
            f"Import {friendly_name} public GPG key to RPM Keyring?",
            selected=Answer.YES,
        )
        == Answer.YES
    ):
        command("rpmkeys --import %s" % path, check=True)


def import_rpm_keys():
    target = config.workdir / "RPM-GPG-KEY-OP5"
    log.debug("Write OP5 GPG KEY to %s", target)
    target.write_text(OP5_RPM_GPGKEY)
    # If key is not already imported, suggest to import it. If user
    # answers no, DNF will prompt to import required keys instead.
    if command("rpm -qi %s" % OP5_RPM_GPGKEY_ID) != 0:
        import_rpm_key_file(target, "OP5")

    # Put to environment to enable use of DNF vars in custom repo configs
    os.environ[DNF_VAR_OP5] = "file://%s" % target


def info(text, *args, always_print=False, **kwargs):
    dlog.info("console: %s", text)
    if config.print_console or always_print:
        print(text, *args, **kwargs)


def init_loggers(conf):
    console = logging.StreamHandler()
    console.setLevel(conf.log_level)
    console.setFormatter(
        fmt=logging.Formatter(
            "{}%(levelname)s:{} {}%(message)s".format(
                conf.color("yellow"),
                conf.color("reset"),
                "[%(name)s] " if conf.debug else "",
            )
        )
    )

    fh = logging.StreamHandler(conf.debug_log_file)
    fh.setFormatter(
        logging.Formatter(
            "***** %(asctime)s %(name)s %(filename)s:%(lineno)s %(funcName)s "
            "%(levelname)s: %(message)s"
        )
    )

    log.setLevel(logging.DEBUG)
    log.addHandler(console)
    log.addHandler(fh)  # include normal logging in debuglog file
    dlog.setLevel(logging.DEBUG)
    dlog.addHandler(fh)
    log.addHandler(fh)
    if conf.debug:
        # with --debug, dump everything to console in addition to file.
        dlog.addHandler(console)


def command(cmd, stdout=False, **kwargs) -> int:
    """
    Run a command.

    Args:
        stdout: If True, child processes inherit parents stdout/stderr,
            otherwise both are written to debug log.
        kwargs: pass to subprocess.run()
    """
    proc = _command(
        cmd,
        stdout=None if stdout else config.debug_log_file,
        stderr=None if stdout else subprocess.STDOUT,
        **kwargs,
    )
    return proc.returncode


def command_output(
    cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs
) -> subprocess.CompletedProcess:
    return _command(cmd, stdout=stdout, stderr=stderr, **kwargs)


def _command(cmd, **kwargs):
    dlog.debug("trace: command=%r kwargs=%s", cmd, kwargs)
    log.debug("Run command: %s", cmd)
    error_hint = kwargs.pop("error_hint", "")
    try:
        p = subprocess.run(
            shlex.split(cmd),
            universal_newlines=True,
            **kwargs,
        )
    except subprocess.CalledProcessError as exc:
        for stream in "stdout", "stderr":
            out = getattr(exc, stream)
            if out:
                for line in out.splitlines():
                    log.error("  %s: %s", stream, line)
        raise OP5Error(
            'Command "%s" failed, exit code: %s. %s',
            cmd,
            exc.returncode,
            error_hint,
            retcode=exc.returncode,
        ) from exc
    except OSError as exc:
        raise OP5Error(
            'Command "%s" failed: (%s). %s',
            cmd,
            exc,
            error_hint,
        ) from exc
    dlog.info("'%s' finished with exit code %d", cmd, p.returncode)
    return p


def sanity_check():
    log.info("Check that we are running with root privileges...")
    if os.getuid() != 0:
        raise OP5Error(
            "Super-user privileges are required to install %s. "
            "Switch to 'root' user or run through 'sudo'.",
            PRODUCT,
        )


def monitor_has_been_installed():
    dlog.debug("Check for traces that OP5 Monitor has been installed before")
    path = pathlib.Path("/opt/monitor/etc")
    if path.exists():
        saved = list(path.glob("*.rpmsave"))
        log.debug("Files in %s: %s", path, saved)
        if saved:
            return True
    path = pathlib.Path("/var/lib/mysql/merlin")
    if path.is_dir():
        log.debug("db directory exists: %s", path)
        return True
    if command("rpm -q %s" % OP5_RPM_GPGKEY_ID) == 0:
        log.debug("OP5 rpm key is installed")
        return True

    return False


def existing_path(input_str, *, is_dir=False) -> pathlib.Path:
    p = pathlib.Path(input_str)
    if is_dir:
        if p.is_dir():
            return p
    elif p.exists():
        return p
    raise argparse.ArgumentTypeError(
        "{} doesn't exist: {}".format("Directory" if is_dir else "File", input_str)
    )


if __name__ == "__main__":
    try:
        main()
    except OP5Error as exc:
        log.error("%s", exc)
        dlog.exception("installation failed")
        log.info("Debug logfile: %s", getattr(config, "debug_log_filename", "n/a"))
        sys.exit(exc.retcode)
    except KeyboardInterrupt:
        info("\n\nAborted.")
        sys.exit(1)
    except subprocess.SubprocessError as exc:
        log.exception("Unexpected external command failure: %s", exc)
        log.error("Debug logfile: %s", getattr(config, "debug_log_filename", "n/a"))
        sys.exit(1)
    except Exception as exc:
        log.info("Got exception", exc_info=True)
        log.critical(
            "Unexpected failure: %s. This is likely a bug, contact %s. "
            "debug log file: %s\n"
            "(running with --verbose can give more details)",
            exc,
            SUPPORT_NAME,
            getattr(config, "debug_log_filename", "N/A"),
        )
        sys.exit(1)
