import json

from flask import Blueprint, jsonify, request, Response, url_for
from oslo_log import log

from nachos.engine.export.config_export import start_export
from nachos.api.v1.schemas import ExportRequest, ExportsFetch
from nachos.api.v1.utilities import parameters, accept, content
from nachos.db import api as db_api


logger = log.getLogger(__name__)
exports = Blueprint('exports', __name__)


@exports.route('/v1/exports/<uuid:export_id>', methods=['GET'])
@parameters()
@accept('application/json')
def get_export(export_id):
    """ Fetch a specific export by id

    :param export_id: String ID of the export
    :return: Response object containing the requested export in JSON format
    """
    logger.debug('Get export')
    sqlalchemy_export = db_api.export_get(export_id)
    return jsonify(sqlalchemy_export.to_response())


@exports.route('/v1/latestexport', methods=['GET'])
@parameters()
@accept('application/json')
def get_latest_export():
    """ Fetch the latest export

    :return: Response object containing the latest export in JSON format
    """
    export = db_api.export_get_latest()

    if not export:
        return jsonify({'name': 'no_exports_found_error',
                        'description': 'No export jobs could be '
                                       'found in the database'}), 404

    return jsonify(export.to_response())


@exports.route('/v1/exports', methods=['GET'])
@parameters(**ExportsFetch.fields)
@accept('application/json')
def get_all_exports():
    """ Fetch multiple exports

    :optional param limit: Int Limit the amount of exports returned. Defaults
                           to None.
    :optional param sort_by: String What property the returned exports will
                             be sorted on. Default creation_time.
    :optional param order: String asc/desc Order the returned exports by
                           ascending or descending order. Default desc.

    :return: Response object containing the list of exports in JSON format
    """
    limit = request.args.get('limit')
    sort_by = request.args.get('sort_by', 'creation_time')
    order = request.args.get('order', 'desc')

    sqlalchemy_exports = db_api.export_get_all(limit, sort_by, order)

    return jsonify([export.to_response() for export in sqlalchemy_exports])


@exports.route('/v1/exports', methods=['POST'])
@parameters()
@content('application/json')
@accept('application/json')
def post_export():
    """ Create a new export

    :param user: String The username of the user that started the export.
    :param export_type: String full_export/user_export What type of export
                        should be performed
    :return: Response object containing the created export in JSON format
    """
    logger.info('Post export')
    export_request = ExportRequest(request.get_json())
    export_request.validate()

    sqlalchemy_export = start_export(export_request)

    response = Response(response=json.dumps(sqlalchemy_export.to_response()),
                        status=201,
                        mimetype="application/json")
    response.headers.add("Location",
                         url_for('.get_export',
                                 export_id=sqlalchemy_export.id))
    return response
