import sys

from flask import jsonify
from oslo_config import cfg
from schematics.exceptions import BaseError as SchematicsBaseError
from schematics.exceptions import FieldError
from sqlalchemy.exc import OperationalError
from werkzeug.http import HTTP_STATUS_CODES
from werkzeug.exceptions import (BadRequest,
                                 NotFound,
                                 UnsupportedMediaType,
                                 NotAcceptable,
                                 InternalServerError)

from nachos.api.v1.exports import exports
from nachos.api.v1.schemas import Error
from nachos.api.v1.utilities import parameters, accept
from nachos.common.consts import ExportStatus
from nachos.common.exceptions import (
    ConfigBackupException,
    ConfigVerificationException,
    ConfigUnknownException,
    NachosError,
    NachosException,
)
from nachos.engine.export.utils import naemon, backup
from nachos.db import api as db_api


def initialize(app, config={}):
    try:
        if fail_unfinished_exports():
            # If there were unfinished exports we try to verify the config
            naemon.verify_configuration(cfg.CONF.monitor.naemon_main_config)

    except (ConfigVerificationException, ConfigUnknownException) as conf_ex:
        # If config verification failed we try to restore the backup
        try:
            backup.restore_backup(
                "{}/backup.tar".format(cfg.CONF.monitor.config_backup_path),
                cfg.CONF.monitor.install_path
            )

        except ConfigBackupException as backup_ex:
            sys.stderr.write(
                "Failed to verify configuration and restore backup:\n"
                "Verification error:\n{}\n"
                "Backup error:\n{}".format(conf_ex, backup_ex)
            )
            sys.exit(1)

    except OperationalError as e:
        sys.stderr.write(
            "Database error. Make sure nachos.cfg "
            "contains correct database credentials\n{}\n".format(e)
        )
        sys.exit(1)

    @app.errorhandler(NotFound)
    @app.errorhandler(BadRequest)
    @app.errorhandler(UnsupportedMediaType)
    @app.errorhandler(InternalServerError)
    @app.errorhandler(NotAcceptable)
    @app.errorhandler(NachosError)
    @app.errorhandler(NachosException)
    def errorhandler(e):

        message = '{} {}'.format(
            e.code, HTTP_STATUS_CODES.get(e.code, 'Unknown Error'))

        return (jsonify([
            Error({
                'name': HTTP_STATUS_CODES.get(e.code),
                'description': message
            }).to_native()
        ]), e.code)

    @app.errorhandler(FieldError)
    def validation_handler(e):
        """Handler for validation and conversion errors"""

        errors = []
        for field_error in e.to_primitive():
            message = '{} {}'.format(
                400, HTTP_STATUS_CODES.get(400, 'Unknown Error'))
            errors.append(Error({
                'name': message,
                'description': field_error
            }))
        return (jsonify([error.to_native() for error in errors]), 400)

    @app.errorhandler(SchematicsBaseError)
    def schematics_handler(e):
        """Schematics errors can contain multiple individual errors."""

        errors = []
        for field, field_errors in e.to_primitive().items():
            if not isinstance(field_errors, list):
                field_errors = [field_errors]
            for field_error in field_errors:
                message = '{} {}'.format(
                    400, HTTP_STATUS_CODES.get(400, 'Unknown Error'))
                errors.append(Error({
                    'name': message,
                    'description': '{}: {}'.format(field, field_error)
                }))
        return (jsonify([error.to_native() for error in errors]), 400)

    @app.after_request
    def format_status_name(response):
        response.status = response.status.lower().title()
        return response

    @app.route('/v1', methods=['GET'])
    @parameters()
    @accept('application/json')
    def get_root():
        endpoints = [{
            'path': str(rule),
            'methods': list(rule.methods)
        } for rule in app.url_map.iter_rules()]
        return jsonify({
            'version': 'v1',
            'endpoints': endpoints
        })

    app.register_blueprint(exports)


def fail_unfinished_exports():
    """ Gets all active exports and set the status to fail. Pending steps
        are set to skipped and running steps are set to fail.

        :returns: True if there are unfinished exports, False otherwise
    """
    failed_exports = False

    unfinished_exports = db_api.export_get_all_unfinished()
    for export in unfinished_exports:
        if export.status not in [ExportStatus.success, ExportStatus.fail]:
            failed_exports = True
            export.status = ExportStatus.fail
            for step, details in export.details.items():
                if details['state'] == ExportStatus.pending.name:
                    details['state'] = ExportStatus.skip.name
                    details['details'] = 'Skipped due to service restart'
                elif details['state'] == ExportStatus.running.name:
                    details['state'] = ExportStatus.fail.name
                    details['details'] = 'Failed due to service restart'
            db_api.export_update(export)

    return failed_exports
