from functools import wraps

from flask import request
from werkzeug.exceptions import BadRequest, NotAcceptable, UnsupportedMediaType


def parameters(**params):
    """
    Validates GET parameters passed to an endpoint, passing parameters not
    defined to an endpoint results in BadRequest, passing an invalid type to a
    parameter results in BadRequest, passing no parameters to an endpoint that
    expects parameters results in BadRequest

    :param **params: All key-value parameters passed.
                     Must be key -> Schematics Type
    """
    def parameters_wrapper(f):
        @wraps(f)
        def parameters_validator(*args, **kwargs):
            values = {}
            if not params and request.args:
                raise BadRequest()
            if params and not set(request.args.keys()).issubset(params):
                raise BadRequest()
            for key, T in params.items():
                value = request.args.get(key, None)
                if T.required or value is not None:
                    values[key] = T.validate(value)
            request.args = values
            return f(*args, **kwargs)
        return parameters_validator
    return parameters_wrapper


def content(*mimes):
    """
    When an endpoint handler is decorated with content the Content-Type of the
    request must match one of the mime-types passed to this Decorator,
    otherwise an UnsupportedMediaType exception is raised.

    :param *mimes: All value parameters passed, expected to be strings
                   representing mime-types.
    """
    def contents_wrapper(f):
        @wraps(f)
        def contents_validator(*args, **kwargs):
            in_mime = request.headers.get('Content-Type', None)
            if (in_mime is None) or (in_mime not in mimes):
                raise UnsupportedMediaType()
            return f(*args, **kwargs)
        return contents_validator
    return contents_wrapper


def accept(*mimes):
    """
    When an endpoint handler is decorated with accept the Accept of the
    request must match one of the mime-types passed to this Decorator,
    otherwise an NotAcceptable exception is raised.

    :param *mimes: All value parameters passed, expected to be strings
                   representing mime-types.
    """
    def accepts_wrapper(f):
        @wraps(f)
        def accepts_validator(*args, **kwargs):
            best = request.accept_mimetypes.best_match(mimes)
            if best is None:
                raise NotAcceptable()
            return f(*args, **kwargs)
        return accepts_validator
    return accepts_wrapper
