"""Routines for configuring nachos."""
import pbr.version
from oslo_config import cfg
from oslo_log import log
from oslo_middleware import cors

# Options under default group.
launch_opt = cfg.ListOpt(
    'server',
    default=['all'],
    help=('Specifies which nachos server to start by the launch script. '
          'Valid options are all or any combination of '
          'api and engine.')
)

api_opts = [
    cfg.HostAddressOpt(
        'host',
        default='127.0.0.1',
        help='nachos API server host'
    ),
    cfg.PortOpt('port', default=8009, help='nachos API server port'),
    cfg.BoolOpt(
        'enable_ssl_api',
        default=False,
        help=('Enable the integrated stand-alone API to service requests'
              'via HTTPS instead of HTTP.')
    ),
    cfg.IntOpt(
        'api_workers',
        help=('Number of workers for nachos API service '
              'default is equal to the number of CPUs available if that can '
              'be determined, else a default worker count of 1 is returned.')
    )
]

engine_opts = [
    cfg.HostAddressOpt(
        'host',
        default='0.0.0.0',
        help=('Name of the engine node. This can be an opaque '
              'identifier. It is not necessarily a hostname, '
              'FQDN, or IP address.')
    ),
    cfg.StrOpt('version',
               default=pbr.version.VersionInfo('nachos').version_string,
               help='The version of the engine.'),
    cfg.IntOpt(
        'engine_workers',
        help=('Number of workers for nachos RPC server '
              'default is equal to the number of CPUs available if that can '
              'be determined, else a default worker count of 1 is returned.')
    )
]

nacoma_opts = [
    cfg.StrOpt(
        'connection',
        default='mysql+pymysql://root@127.0.0.1:3306/nacoma',
        help='The URL to the Nacoma database.'
    ),
    cfg.StrOpt(
        'blob_encoding',
        default='utf-8',
        help=(
            "The encoding used in Nacoma binary data columns, such as "
            "changelog state columns or command command_line columns."
        )
    ),
]

monitor_opts = [
    cfg.StrOpt(
        'install_path',
        default='/opt/monitor/',
        help='Path to the Monitor installation.'
    ),
    cfg.StrOpt(
        'naemon_main_config',
        default='/opt/monitor/etc/naemon.cfg',
        help='Path to the main Naemon configuration file.'
    ),
    cfg.StrOpt(
        'config_backup_path',
        default='/var/cache/naemon/',
        help='Path to directory where to save backed up configurations'
    ),
    cfg.IntOpt(
        'days_to_keep_backups',
        default=10,
        help='Number of days to keep backed up configurations'
    ),
    cfg.StrOpt(
        'save_hook_path',
        default='/opt/monitor/op5/nacoma/hooks/save/',
        help='Path to directory where the save hooks are located'
    ),
    cfg.BoolOpt(
        'verify_before_commit',
        default=False,
        help=('Reduce downtime in case of error when verifying config'
              'by sacrificing config export speed')
    ),
    cfg.StrOpt(
        'config_encoding',
        default='utf-8',
        help="The encoding used in the Naemon config files."
    ),
]

CONF = cfg.CONF

API_GROUP = 'api'
ENGINE_GROUP = 'engine'
NACOMA_GROUP = 'nacoma'
MONITOR_GROUP = 'monitor'

CONF.register_opts(api_opts, group=API_GROUP)
CONF.register_opts(engine_opts, group=ENGINE_GROUP)
CONF.register_opts(nacoma_opts, group=NACOMA_GROUP)
CONF.register_opts(monitor_opts, group=MONITOR_GROUP)

CONF.register_cli_opts([
    launch_opt
])

_DEFAULT_LOG_LEVELS = [
    'eventlet.wsgi.server=WARN',
    'oslo_service.periodic_task=INFO',
    'oslo_service.loopingcall=INFO',
    'oslo_db=WARN'
]


def list_opts():
    yield API_GROUP, api_opts
    yield ENGINE_GROUP, engine_opts
    yield NACOMA_GROUP, nacoma_opts
    yield MONITOR_GROUP, monitor_opts


def parse_args(args=None, usage=None, default_config_files=None):
    default_log_levels = log.get_default_log_levels()
    default_log_levels.extend(_DEFAULT_LOG_LEVELS)
    log.set_defaults(default_log_levels=default_log_levels)

    log.register_options(CONF)

    CONF(
        args=args,
        project='nachos',
        version=pbr.version.VersionInfo('nachos').version_string,
        usage=usage,
        default_config_files=default_config_files
    )


def set_config_defaults():
    """This method updates all configuration default values."""
    set_cors_middleware_defaults()


def set_cors_middleware_defaults():
    """Update default configuration options for oslo.middleware."""
    cors.set_defaults(
        allow_methods=['GET',
                       'PUT',
                       'POST',
                       'DELETE',
                       'PATCH']
    )
