import sys
import threading
from datetime import datetime

from oslo_db.sqlalchemy import enginefacade

from nachos.api.v1.schemas import ExportDetails
from nachos.db.sqlalchemy.models import Export
from nachos.db.sqlalchemy import migration
from nachos.common import exceptions as exception
from nachos.common.consts import ExportStatus, ExportType

_main_context_manager = None
_CONTEXT = threading.local()


def _get_main_context_manager():
    global _main_context_manager
    if not _main_context_manager:
        _main_context_manager = enginefacade.transaction_context()
    return _main_context_manager


def get_engine():
    return _get_main_context_manager().get_legacy_facade().get_engine()


def get_backend():
    """ The backend is this module itself. """
    return sys.modules[__name__]


def read_session():
    return _get_main_context_manager().reader.using(_CONTEXT)


def write_session():
    return _get_main_context_manager().writer.using(_CONTEXT)


def db_sync(engine, version=None):
    """ Migrate the database to `version` or the most recent version. """
    return migration.db_sync(engine, version=version)


def db_version(engine):
    """ Display the current database version. """
    return migration.db_version(engine)


def export_get(export_id):
    """ Get a specific export by ID """
    with read_session() as session:
        query = session.query(Export)
        export = query.filter(Export.id == export_id)
        obj = export.one_or_none()

        if not obj:
            raise exception.DBEntityNotFoundError()
        return obj


def export_get_all(limit=None, sort_by='creation_time', order='desc'):
    """ Get all exports

        Args:
            limit (int): Limits the amount of returned exports
            sort_by (string): Property the exports are sorted by
            order (string): Order by asc or desc

    """
    with read_session() as session:

        sort = getattr(Export, sort_by)
        order = getattr(sort, order)

        if limit:
            query = session.query(Export).order_by(order()).limit(limit)
        else:
            query = session.query(Export).order_by(order())

        exports = query.all()

        return exports


def export_get_all_unfinished():
    """ Get all unfinished exports """
    with read_session() as session:
        exports = session.query(Export).filter(
            Export.status.notin_([ExportStatus.success,
                                  ExportStatus.fail])).all()
        return exports


def export_get_latest():
    """ Get the newest export by creation_time """
    with read_session() as session:
        latest_export = session.query(Export).order_by(
            Export.creation_time.desc()).first()
        return latest_export


def export_create(values):
    """ Create a new export entry

        Args:
            values (ExportRequest): ExportRequest schema object

        Returns:
            export (Export): Export schema object

    """
    with write_session() as session:
        export = Export(user=values.user,
                        creation_time=datetime.now(),
                        status=ExportStatus.pending,
                        export_type=ExportType[values.export_type],
                        details=ExportDetails().to_primitive(),
                        error=None)
        session.add(export)
        return export


def export_update(values):
    """ Update an existing export

        Args:
            values (Export): Export schema object

    """
    assert isinstance(values, Export)
    with write_session() as session:
        session.query(Export).filter(Export.id == values.id).update(
            {
                Export.creation_time: values.creation_time,
                Export.status: values.status,
                Export.user: values.user,
                Export.export_type: values.export_type,
                Export.error: values.error,
                Export.details: values.details
            })


def export_delete(values):
    """ Delete an existing export

        Args:
            values (Export): Export schema object

    """
    assert isinstance(values, Export)
    with write_session() as session:
        session.delete(values)
