import enum
import uuid

from sqlalchemy import (Table, Column, MetaData,
                        DateTime, String, Enum)
from sqlalchemy.types import TypeDecorator, CHAR


meta = MetaData()


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return uuid.UUID(value)


class ExportStatus(enum.Enum):
    pending = 1
    running = 2
    success = 3
    fail = 4


class ExportType(enum.Enum):
    full_export = 1
    user_export = 2


exports = Table(
    'exports', meta,
    Column('id', GUID, nullable=False, primary_key=True),
    Column('creation_time', DateTime),
    Column('status', Enum(ExportStatus), nullable=False),
    Column('user', String(255), unique=False, nullable=False),
    Column('export_type', Enum(ExportType), nullable=False),
    mysql_engine='InnoDB',
    mysql_charset='utf8'
)


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    exports.create()


def downgrade(migrate_engine):
    meta.bind = migrate_engine
    exports.drop()
