import enum
import uuid
import json

from sqlalchemy import (Table, Column, MetaData,
                        DateTime, String, Enum)
from sqlalchemy.types import TypeDecorator, CHAR, VARCHAR


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::

        JSONEncodedDict(255)

    """

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return uuid.UUID(value)


class ExportStatus(enum.Enum):
    pending = 1
    running = 2
    success = 3
    fail = 4


class ExportType(enum.Enum):
    full_export = 1
    user_export = 2


def upgrade(migrate_engine):
    meta = MetaData(migrate_engine)
    exports = Table(
        'exports', meta,
        Column('id', GUID, nullable=False, primary_key=True),
        Column('creation_time', DateTime),
        Column('status', Enum(ExportStatus), nullable=False),
        Column('user', String(255), unique=False, nullable=False),
        Column('export_type', Enum(ExportType), nullable=False),
        Column('details', JSONEncodedDict(1023), nullable=False),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )
    error = Column('error', JSONEncodedDict(1023), nullable=True)
    exports.create_column(error)


def downgrade(migrate_engine):
    meta = MetaData(migrate_engine)
    exports = Table(
        'exports', meta,
        Column('id', GUID, nullable=False, primary_key=True),
        Column('creation_time', DateTime),
        Column('status', Enum(ExportStatus), nullable=False),
        Column('user', String(255), unique=False, nullable=False),
        Column('export_type', Enum(ExportType), nullable=False),
        Column('details', JSONEncodedDict(1023), nullable=False),
        Column('error', JSONEncodedDict(1023), nullable=True),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )
    exports.drop_column('error')
