"""
SQLAlchemy models for nachos data.
"""

import uuid

from sqlalchemy import Column, String, DateTime, Enum
from sqlalchemy.ext.declarative import declarative_base

from nachos.common.consts import ExportStatus, ExportType, ExportSteps
from nachos.db.sqlalchemy.types import GUID, JSONEncodedDict


BASE = declarative_base()


class Export(BASE):
    __tablename__ = "exports"

    id = Column(GUID, primary_key=True, default=uuid.uuid4, nullable=False)
    creation_time = Column(DateTime)
    status = Column(Enum(ExportStatus), nullable=False)
    user = Column(String(255), unique=False, nullable=False)
    export_type = Column(Enum(ExportType), nullable=False)
    details = Column(JSONEncodedDict(1023), nullable=False)
    error = Column(JSONEncodedDict(1023), nullable=True)

    allowed_steps = [
        ExportSteps.backup.name,
        ExportSteps.commit.name,
        ExportSteps.verify.name,
        ExportSteps.reload.name,
        ExportSteps.rollback.name,
    ]
    allowed_keys = ["state", "details"]

    def get_detail(self, step, key, default=''):
        """Gets nested step progress

        :param step: Step name
        :param key: Key name
        :param default='': Default return value
        """
        return self.details.get(step, {}).get(key, default)

    def set_detail(self, step, key, value=''):
        """Sets nested step progress

        :param step: Step name
        :param key: Key name
        :param value: Nested key value
        """

        assert key in self.allowed_keys
        assert step in self.allowed_steps

        self.details[step][key] = value

    def skip_pending(self):
        details = self.details
        for step in self.details:
            if details[step].get('state') == ExportStatus.pending.name:
                details[step]['state'] = ExportStatus.skip.name

    def set_error_from_exception(self, exception):
        self.error = {
            'name': type(exception).__name__,
            'description': str(exception)
        }

    def to_response(self):
        return {
            'id': str(self.id),
            'user': self.user,
            'export_type': self.export_type.name,
            'status': self.status.name,
            'creation_time': self.creation_time.strftime(
                '%Y-%m-%dT%H:%M:%S.%f%z'
            ),
            'error': self.error,
            'status_details': [
                {
                    'state': state,
                    'progress': int(
                        self.get_detail(detail, 'state') == 'success'
                    ),
                    'details': self.get_detail(detail, 'details')
                }
                for state, detail in (
                    ('backup', 'backup'),
                    ('config_generation', 'commit'),
                    ('verification', 'verify'),
                    ('commit', 'reload')
                )
            ]
        }
