import uuid
import datetime
import time
import json

from sqlalchemy.types import TypeDecorator, CHAR, Integer, VARCHAR


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return uuid.UUID(value)


class Epoch(TypeDecorator):
    """ a type that decorates DateTime, converts to unix time on
    the way in and to datetime.datetime objects on the way out.

    Credits to Pavel Repin on Stackoverflow
    """
    impl = Integer
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        if isinstance(value, datetime.datetime):
            return int(time.mktime(value.timetuple()))
        raise ValueError("Not a valid datetime: {0}".format(
            type(value).__name__))

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        else:
            return datetime.datetime.fromtimestamp(float(value))


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::

        JSONEncodedDict(255)

    """

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value
