import codecs
import functools
import os
import subprocess
import pathlib
import tempfile
import shutil
from collections import OrderedDict
from multiprocessing import Process
from itertools import chain

from oslo_config import cfg
from oslo_log import log
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from nachos.common.consts import ExportStatus, ExportType
from nachos.common.exceptions import (
    ConfigExportException,
    ConfigSyntaxException,
    JobAlreadyRunningError,
)
from nachos.db import api as db_api
from nachos.engine.export.utils.backup import (
    create_backup,
    restore_backup,
    archive_backup,
)
from nachos.engine.export.writers.helpers import write_header
from nachos.engine.export.utils.naemon import (verify_configuration,
                                               naemon_reload)
from nachos.engine.export.utils.config_parser import (
    parse_definitions,
    parse_config_key,
)
from nachos.engine.export.utils.updates_helper import (
    Updates,
    ChangeStatus,
)
from nachos.engine.export.writers.writemapper import write_object
from nachos.nacoma import Nacoma
from nachos.nacoma.exceptions.exceptions import NacomaLockAquireFail

from nachos.engine.export.utils.config_key_helper import get_config_key


logger = log.getLogger(__name__)
export_process = None


def start_export(export_request):
    """ Spawns a subprocess responsible for the configuration export.

    :param export_request: ExportRequest object containing the POST data
    :return: The SQLALchemy ORM object of the current export
    """
    global export_process
    if export_process is not None and export_process.is_alive():
        raise JobAlreadyRunningError

    sqlalchemy_export = db_api.export_create(export_request)

    args = [sqlalchemy_export, cfg.CONF.nacoma.connection, export_request.user]
    export_process = Process(target=_execute_export_steps, args=args)
    export_process.start()

    return sqlalchemy_export


def _execute_export_steps(sqlalchemy_export, nacoma_url, username):
    """ Executes the configuration backup, writing, restoration and reloads
        Naemon.

        Here we execute most of the steps and functions related to a
        configuration export. A Nacoma database session is is created and
        injected into the Nacoma object. A Nacoma database lock token is
        inserted to the database so the UI won't allow object modification or
        concurrent exports.

        In the first step all Naemon configuration files listen in the Nacoma
        file_tbl are backed up so the configuration can be restored in case of
        an error. If user_export is specified then the database will be rolled
        back to its initial state using the changelog entries, and the
        changelog entries owned by the exporting user will be applied to the
        database, so only the user specific changes are exported.

        The objects are then written to the configuration files,
        and Naemon is reloaded. Post export the save hooks are executed,
        the changelog entries are moved to changelog_history, and an
        action_log entry is created so Nacoma won't try to reimport the newly
        created configuration.

        If an exceptions occur during any step the configuration backup will
        be restored, and export step set to fail.

    :param sqlalchemy_export: The SQLALchemy ORM object of the current export
    :param nacoma_url: String The Nacoma database URL
    :param username: String The user who initiated the export
    """
    assert nacoma_url is not None, 'execute_exports called without nacoma_url'
    assert username is not None, 'execute_export called without username'

    engine = create_engine(nacoma_url)
    Session = sessionmaker(bind=engine)
    session = Session()

    try:
        nacoma = Nacoma()
        nacoma.set_session(session)
        nacoma.aquire_database_lock(username=username)
    except NacomaLockAquireFail as e:
        _fail_export(sqlalchemy_export, e, update=True)
        return

    backup_path = None

    try:
        prepared_backup = functools.partial(
            _create_backup,
            nacoma
        )

        logger.info("Backing up config")
        backup_path = _run_step(
            sqlalchemy_export,
            prepared_backup,
            'backup',
            'Failed to back up configuration'
        )

        user_export = sqlalchemy_export.export_type == ExportType.user_export
        more_than_one_user = _more_than_one_user(nacoma.changelogs)
        prepared_export = functools.partial(
            _write_objects,
            nacoma,
            sqlalchemy_export,
            user_export and more_than_one_user,
            username,
        )

        logger.info('Exporting config')
        write_changelog_ok = _run_step(
            sqlalchemy_export,
            prepared_export,
            'commit',
            'Failed to export database due to {}. '
            'Please view the logs for more detail.'
        )

        logger.info("Verifying config")
        _run_step(
            sqlalchemy_export,
            _verify_config,
            'verify',
            'Failed to verify config due to {}. '
            'Please view the logs for more detail'
        )

        logger.info("Reloading Naemon")
        _run_step(
            sqlalchemy_export,
            naemon_reload,
            'reload',
            'Failed to reload Naemon due to {}. '
            'Please view the logs for more detail'
        )

        _execute_save_hooks(nacoma.changelogs, username)

        if user_export:
            nacoma.archive_changes(username)
        else:
            nacoma.archive_changes()

        if user_export and more_than_one_user and not write_changelog_ok:
            nacoma.apply_changelog()

        archive_backup(backup_path, cfg.CONF.monitor.days_to_keep_backups)

        nacoma.log_action('export', username)
        logger.info('Configuration successfully exported')

    except Exception as e:
        _fail_export(sqlalchemy_export, e)
        _rollback(
            session, sqlalchemy_export, backup_path,
            cfg.CONF.monitor.install_path
        )
    else:
        sqlalchemy_export.status = ExportStatus.success
    finally:
        nacoma.release_database_lock()
        # If any steps are still pending, they are to be considered skipped
        sqlalchemy_export.skip_pending()
        db_api.export_update(sqlalchemy_export)


def _write_objects(nacoma, sqlalchemy_export, user_export=False, username=None):
    """ Writes all changed objects to config files. The primary strategy is to
        only write the objects that's been changed, but not all cases are
        implemented and we are forced to fall back to the slower process of
        writing all objects.

    :param nacoma: API to the Nacoma database
    :param user_export: Boolean Only changes made by `username` are written, if True
    :param username: String The user who initiated the export
    """
    if user_export and not username:
        raise AssertionError("Required username missing for user_export")
    try:
        updates = _retrieve_merge_and_map_changes(nacoma, user_export, username)
        _write_updates(updates)
        return True
    except NotImplementedError as e:
        logger.info("Falling back to full database export: {}".format(str(e)))
        sqlalchemy_export.set_detail(
            'commit',
            'details',
            value=(
                '{}. Because of this, all objects must be exported, making this '
                'export take longer time than normal.'
            ).format(str(e))
        )
        db_api.export_update(sqlalchemy_export)
        _write_database_objects(nacoma, user_export, username)
        return False


def _retrieve_merge_and_map_changes(nacoma, user_export, username):
    """ Retrieves, merges and maps changes to files

        Returns a dictionary of filenames mapped to the objects that reside in
        those files. The objects are split up into two categories, "add" and
        "change", which represent objects which are new and should be added
        (add), and objects which should be updated or deleted (change).

    :param nacoma: API to the Nacoma database
    :param user_export: Boolean Only changes made by `username` are retrieved, if True
    :param username: String The user who initiated the export
    :return: updates
    """

    changes = _generate_changes(nacoma, user_export, username)
    changes = _merge_changes(changes)
    updates = _map_changes_to_files(nacoma, changes)
    return updates


def _generate_changes(nacoma, user_export, username):
    """ Generates all changelog entries from Nacoma, plus changelog entries for
        any dependent objects.

    :param nacoma: API to the Nacoma database
    :param user_export: Boolean Only changes made by `username` are retrieved, if True
    :param username: String The user who initiated the export
    :yields: Changelog entries for all changed objects
    """

    if user_export:
        # In a user export we need to include all changelog entries for all
        # objects changed by the responsible user, and not just the entries made
        # by that user.
        changed_objects = {
            change.object_identifier
            for change in nacoma.changelogs_by_username(username)
        }
    else:
        changed_objects = None

    for change in nacoma.changelogs:
        if changed_objects and change.object_identifier not in changed_objects:
            continue
        yield change
        if change.object_type in ("hostgroup", "servicegroup", "contactgroup"):
            all_members = change.get_all_state_members("members")
            if change.object_type == "hostgroup":
                for obj_id in all_members:
                    obj = nacoma.host_by_id(obj_id)
                    if obj:
                        yield obj.generate_changelog()
            elif change.object_type == "servicegroup":
                for obj_id in all_members:
                    obj = nacoma.service_by_id(obj_id)
                    if obj:
                        yield obj.generate_changelog()
            elif change.object_type == "contactgroup":
                for obj_id in all_members:
                    obj = nacoma.contact_by_id(obj_id)
                    if obj:
                        yield obj.generate_changelog()
        elif change.object_type in ("host", "service"):
            all_children = change.get_all_state_members("children")
            for obj_id in all_children:
                obj = nacoma.object_by_type_and_id(change.object_type, obj_id)
                if obj:
                    yield obj.generate_changelog()


def _merge_changes(changes):
    """ Pre-process the changes and return one changelog entry per object, that
        holds the original and final values that we care about.

    :param changes: List of changelog objects
    """
    merged_changes = OrderedDict()
    for change in changes:
        key = change.object_identifier
        merged_change = merged_changes.get(key)
        if merged_change:
            merged_change.merge_change(change)
        else:
            merged_changes[key] = change
    return merged_changes.values()


def _map_changes_to_files(nacoma, changes):
    """ Break down the changes so that we know what objects to add, update and
        delete for each config file.

    :param nacoma: API to the Nacoma database
    :param changes: List of changelog objects
    """
    updates = Updates()
    for change in changes:
        if change.is_deleted:
            if change.is_new:
                # It's never been written to file, so we just skip it.
                continue
            if change.object_type in ("contact", "contactgroup", "timeperiod"):
                # Deleting some objects do not generate changelog entries for
                # objects references them. This is true for multiple contact
                # and contactgroup relations, and for timeperiods that are
                # excluded by other timeperiods. Without state information or
                # changelog entries, we have no way of knowing about these
                # relations. So without a fix in Nacoma we are forced to do a
                # full database export in these cases.
                raise NotImplementedError(
                    "One or more {}s have been deleted".format(change.object_type)
                )
            config_key = get_config_key(change, changes, nacoma)
            filename = nacoma.file_by_id(change.old_file_id).name
            updates.deleted(filename, config_key)
        else:
            mapper = nacoma.object_by_type_and_id(
                change.object_type,
                change.object_id
            )
            filename = mapper.file.name
            if change.is_new:
                updates.new(filename, mapper)
            else:
                config_key = get_config_key(change, changes, nacoma, mapper=mapper)
                if change.is_renamed:
                    raise NotImplementedError("One or more objects have been renamed")
                if change.is_moved:
                    old_filename = nacoma.file_by_id(change.old_file_id).name
                    new_filename = nacoma.file_by_id(change.new_file_id).name
                    updates.deleted(old_filename, config_key)
                    updates.new(new_filename, mapper)
                else:
                    updates.changed(filename, config_key, mapper)
    return updates


def _write_updates(updates):
    """ Writes all updates to file.

    :param updates: Updates to be written to file
    """

    encoding = cfg.CONF.monitor.config_encoding

    # Update each config file that has changes
    for filename, (new, changed) in updates.items():
        filepath = os.path.join(cfg.CONF.monitor.install_path, filename)
        # Append new objects to the end of the config file.
        if new:
            with open(filepath, "ab") as outfile:
                encoded_outfile = codecs.getwriter(encoding)(outfile)
                for mapper in new:
                    encoded_outfile.write("\n")
                    write_object(mapper, encoded_outfile)
        # Parse the config file looking for objects to change or delete.
        if changed:
            _write_changed_objects(changed, filepath)
            _check_all_done(changed, filename)


def _check_all_done(changed, filename):
    not_done = [
        config_key
        for config_key, status in changed.items()
        if status is not ChangeStatus.DONE
    ]
    if not_done:
        for config_key in not_done:
            logger.warning("Object could not be found in {}: {}".format(
                filename, config_key
            ))
        raise NotImplementedError(
            "One or more object definitions could not be located "
            "(see the log for details)"
        )


def _write_changed_objects(changed, filepath):
    """ Copy the config file object by object to a temporary file. Objects
        that we identify as changed are instead written from the database
        object, and deletions are not written at all. Finally we replace the
        config file with the temporary file.

    :param changed: List of config_keys and mapper objects to write
    :param filepath: Path to the config file where the objects are saved
    """
    outfile = tempfile.NamedTemporaryFile(mode="wb", delete=False)
    p_linenum = 0
    encoding = cfg.CONF.monitor.config_encoding
    try:
        with open(filepath, "rb") as infile:
            encoded_outfile = codecs.getwriter(encoding)(outfile)
            for p_ws, p_definition, p_linenum in parse_definitions(infile):
                if p_definition is None:
                    outfile.write(b"".join(p_ws))
                    break
                config_key = parse_config_key(p_definition, encoding=encoding)
                change = changed.get(config_key)
                # At this point change can be one of four different types: None,
                # ChangeStatus.DONE, ChangeStatus.DELETED or a db mapper object.
                # None means we don't have a changelog entry for the parsed
                # object, ChangeStatus.DONE means we've already processed this
                # changelog entry on a previously parsed object,
                # ChangeStatus.DELETED means the parsed object should be deleted,
                # and an object mapper means we should use that to write an
                # updated definition.
                if change is None:
                    # Just copy the object definition.
                    outfile.write(b"".join(p_ws))
                    outfile.write(b"".join(p_definition))
                elif change is ChangeStatus.DONE:
                    raise NotImplementedError(
                        "One or more {}s match multiple object definitions".format(
                            config_key[0]
                        )
                    )
                else:
                    if change is not ChangeStatus.DELETED:
                        # Remove any preceding comment, it will be rewritten.
                        if p_ws and p_ws[-1].lstrip().startswith(b"#"):
                            p_ws = p_ws[:-1]
                        outfile.write(b"".join(p_ws))
                        # Now change must be a mapper object, let's write it.
                        write_object(change, encoded_outfile)
                    changed[config_key] = ChangeStatus.DONE
        outfile.close()
        os.chmod(outfile.name, 0o660)
        shutil.move(outfile.name, filepath)
    except (ConfigSyntaxException, UnicodeDecodeError) as e:
        raise ConfigSyntaxException("Failed parsing object definition in {}: {}".format(
            "{} line {}".format(filepath, p_linenum) if p_linenum else filepath,
            e,
        ))
    finally:
        try:
            os.remove(outfile.name)
        except FileNotFoundError:
            pass


def _write_database_objects(nacoma, user_export, username):
    """ Opens configuration file handles and write the Nacoma objects to disk

    :param nacoma: API to the Nacoma database
    :param user_export: Boolean Only changes made by `username` are written, if True
    :param username: String The user who initiated the export
    """

    if user_export:
        nacoma.rollback_changelog()
        nacoma.apply_changelog(username)

    objects = chain(nacoma.commands, nacoma.contactgroups, nacoma.contacts,
                    nacoma.hostdependencies, nacoma.hostescalations,
                    nacoma.hostgroups, nacoma.hosts,
                    nacoma.servicedependencies, nacoma.serviceescalations,
                    nacoma.servicegroups, nacoma.services,
                    nacoma.timeperiods)
    handles = {}

    try:
        for f in nacoma.files:
            filepath = os.path.join(cfg.CONF.monitor.install_path, f.name)
            handles[f.name] = open(
                filepath, 'w', encoding=cfg.CONF.monitor.config_encoding
            )
            write_header(handles[f.name], username)
        for obj in objects:
            f = getattr(obj, 'file', None)
            if f:
                write_object(obj, handles[f.name])
    except Exception:
        raise
    finally:
        for handle in handles.values():
            handle.close()


def _rollback(session, sqlalchemy_export, backup_path, backup_dest):
    """ Rollback configuration, naemon and the session

    :param session: SQLAlchemy session to the Nacoma database
    :param details: ExportDetails object
    :param backup_path: String Path to the Naemon configuration backup file
    :param backup_dest: String path to the Naemon configuration folder
    """
    sqlalchemy_export.set_detail(
        'rollback',
        'state',
        value=ExportStatus.running.name
    )
    try:
        restore_backup(backup_path, backup_dest)
        session.rollback()
        naemon_reload()
    except (OSError, ConfigExportException) as e:
        logger.critical('Rollback failed: {}'.format(e.message))
        sqlalchemy_export.set_detail(
            'rollback',
            'state',
            value=ExportStatus.fail.name
        )
        sqlalchemy_export.set_detail(
            'rollback',
            'details',
            value='Failed to reload naemon in rollback'
        )
    else:
        sqlalchemy_export.set_detail(
            'rollback',
            'state',
            value=ExportStatus.success.name
        )
    finally:
        db_api.export_update(sqlalchemy_export)


def _fail_export(sqlalchemy_export, exception, update=False):
    """ Log an exception and update export status.

    :param sqlalcemy_export: Export DB object
    :param exception: The exception raised
    :param update: Set to True to write the DB updates immediately
    """
    # Do not log traceback of ConfigExportException
    logger.error(
        exception,
        exc_info=not isinstance(exception, ConfigExportException)
    )
    sqlalchemy_export.set_error_from_exception(exception)
    sqlalchemy_export.status = ExportStatus.fail
    if update:
        db_api.export_update(sqlalchemy_export)


def _create_backup(nacoma):
    """ Queries nacoma for a list of config files and creates a backup from
        them.

    :param nacoma: Nacoma database session object.
    :return: The path to the backup
    """
    files = [f.name for f in nacoma.files]
    backup_path = create_backup(
        cfg.CONF.monitor.install_path,
        files,
        cfg.CONF.monitor.config_backup_path
    )
    return backup_path


def _verify_config():
    """ Verifies the exported configuration based on the `verify_before_commit`
        config option.
    """
    if cfg.CONF.monitor.verify_before_commit is True:
        verify_configuration(cfg.CONF.monitor.naemon_main_config)
    else:
        logger.debug('Skipping configuration verification')


def _run_step(sqlalchemy_export, func, step, error_msg):
    """ Runs a function in the context of a sqlalchemy export.

    :param sqlalchemy_export: The database model of an export
    :param func: The function to run
    :param step: The name of the step which is being executed
    :param error_msg: The error message to throw in case of failure
    """
    sqlalchemy_export.set_detail(
        step,
        'state',
        value=ExportStatus.running.name
    )
    db_api.export_update(sqlalchemy_export)
    try:
        ret_val = func()
    except Exception as e:
        sqlalchemy_export.set_detail(
            step,
            'state',
            value=ExportStatus.fail.name
        )
        sqlalchemy_export.set_detail(
            step,
            'details',
            value=error_msg.format(type(e).__name__)
        )
        raise
    else:
        sqlalchemy_export.set_detail(
            step,
            'state',
            value=ExportStatus.success.name
        )
        return ret_val
    finally:
        db_api.export_update(sqlalchemy_export)


def _execute_save_hooks(changelogs, username, full_name="unknown"):
    """ Executes all save hooks

    Executes every executable file in the the save hook directory
    with formatted changelogs piped to their stdin streams.

    full_name is not yet implemented as it was in Nacoma,
    and will default to unknown.

    :param changelogs: A Generator for ChangelogMapper objects
    :param username: Username of the person executing the command
    :param full_name: Full name of the person according to contact_name.
    :returns: None
    """

    hook_dir = pathlib.Path(cfg.CONF.monitor.save_hook_path)
    if not hook_dir.is_dir():
        logger.debug("Save hook directory not found")
        return

    # gen(changelog) -> [str]
    formatted_entries = list(map(_format_changelog_entry, changelogs))

    for hook in hook_dir.iterdir():
        if not _is_executable(str(hook)) or not hook.is_file():
            continue

        logger.debug("Executing hook %s", hook)

        try:
            with subprocess.Popen(
                    [str(hook), username, full_name],
                    stdin=subprocess.PIPE
                    ) as proc:
                for entry in formatted_entries:
                    proc.stdin.write(bytes(entry, "utf-8"))
        except OSError as ex:
            logger.error(
                "Execution failed for {}, got exception: {}".format(hook, ex)
            )


def _is_executable(file_path):
    """ Checks if a file is executable

    :param file_path: A file path as a string
    :returns: True if executable, False otherwise
    """
    return os.access(file_path, os.X_OK)


def _format_changelog_entry(entry):
    """ Formats a changelog entry

    The required format is a string of tab seperated key-value pairs structured
    as key='value'.

    The key value pairs are as follows:
        object_type='<string>'"
        object_id='<int>'
        oldname='<string>'"
        newname='<string>'"
        username='<string>'
        time='<string>'
        pre='k1=v1;k2=v2;kn=vn'
        post='k1=v1;k2=v2;kn=vn'

    Legend:
    * object_type: host, service, hostgroup, etc.
    * object_id is the id in the object database before the logged action
      was applied (ie, a deleted object will still have an object id).
    * oldname: The Nagios name of the object before the action.
    * newname: The Nagios name of the object after the action.
    * username: The name of the user who performed the action.
    * time: Timestamp of the logged action, e.g. 2018-11-28 09:26:44
    * pre: The changed keys in the object as they were before the change
    * post: The changed keys in the object as they were after the change

    If an object was deleted, "newname" and "post" will be nul strings.
    If an object was created, "oldname" and "pre" will be nul strings.

    :param entry: A ChangelogMapper object
    :returns: A formatted changlog string
    """
    formatted_entry = (
        "object_type='{}'\t"
        "object_id='{}'\t"
        "oldname='{}'\t"
        "newname='{}'\t"
        "username='{}'\t"
        "time='{}'\t"
        "pre='{}'\t"
        "post='{}'\n"
    ).format(
        entry.object_type,
        entry.object_id,
        entry.oldname if entry.oldname else "",
        entry.newname if entry.newname else "",
        entry.username,
        entry.time.strftime("%Y-%m-%d %H:%M:%S"),
        entry.oldstate.decode(cfg.CONF.nacoma.blob_encoding) if entry.oldstate else "",
        entry.newstate.decode(cfg.CONF.nacoma.blob_encoding) if entry.newstate else "",
    )
    return formatted_entry


def _more_than_one_user(changelogs):
    """ Checks if changes have been made by more than one user.

    :param changelogs: The changelogs to iterate over
    :returns: True if more than one user, False otherwise
    """
    try:
        first_user = next(changelogs).username
    except StopIteration:
        # Raised by the call to __next__() if there were no changelogs
        return False

    for change in changelogs:
        if change.username != first_user:
            return True
    return False
