import datetime
import os
import pathlib
import tarfile
import time

from nachos.common.exceptions import (
    ConfigBackupException,
    ArchiveBackupException,
)


def create_backup(root_dir, backup_list, backup_dest):
    """
    Generates a tar file from a list of files.
    :param root_dir: path to where to start backing up files from
    :param backup_list: list of file paths relative to root_dir
    :param backup_dest: path to where to create the archive
    :return: path to the created archive
    """
    assert root_dir, 'No root dir provided'
    assert backup_dest, 'Invalid path to backup destination given'

    if not backup_list:
        raise ConfigBackupException('Backup file list was empty')

    backup_path = os.path.join(backup_dest, "backup.tar")

    with tarfile.TarFile(backup_path, mode="w") as archive:
        for file in backup_list:
            archive.add(os.path.join(root_dir, file), arcname=file)

    return backup_path


def restore_backup(backup_file, dest):
    """
    Unpack a tar file with backed up files.

    :param backup_file: path to backup file
    :param dest: where to unpack the files, typically the monitor install path
    :return: None
    """

    assert backup_file, 'No backup file provided'
    assert dest, 'No destination given for backup restore'

    backup_path = pathlib.Path(backup_file)
    if not backup_path.exists():
        # Get the latest archived backup
        backup_file = _get_latest_backup(str(backup_path.parent))

    with tarfile.open(backup_file, "r:") as tar:
        tar.extractall(dest)


def archive_backup(backup_path, days_to_keep=10):
    """ Renames the latest created backup with current timestamp

    After a successful export the backup is archived and saved in
    the backup directory so that if the configuration files were
    somehow corrupted without Nachos catching the issue there
    are backups which can be used to restore. Since this will
    create a new backup for every export we also go through
    and remove every backup older than 7 days as a last step.

    The process of archiving is just renaming the backup file
    to the current unix timestamp.
    Example: 1551791533.tar
    Which is: 03/05/19 - 14:12:13

    :param backup_path: Path to the backup to archive
    :param days_to_keep: Number of days to keep an archived backup
    :return: None
    """

    path = pathlib.Path(backup_path)
    if not path.exists():
        raise ArchiveBackupException(
            "Archive backup called with non-existent path: {}"
            .format(backup_path)
        )

    filename = '{:.0f}.tar'.format(time.time())
    archive_path = pathlib.Path(path.parent / filename)
    path.rename(archive_path)

    _prune_backups(path.parent, days_to_keep)


def _get_latest_backup(backup_path):
    """ Returns the most recent archived backup

    :return: The path to the latest backup
    """
    sorted_backups = _get_sorted_backups(backup_path)
    if not sorted_backups:  # Empty list
        raise ConfigBackupException(
            "Could not restore since backup dir {} was empty"
            .format(backup_path)
        )
    return str(sorted_backups[-1])


def _get_sorted_backups(backup_path):
    """ Returns a sorted list of backups
    """
    backup_dir = pathlib.Path(backup_path)
    return sorted(backup_dir.glob('*.tar'))


def _prune_backups(backup_path, days):
    """ Removes all backups older than "days" days
    """
    limit = datetime.datetime.now() - datetime.timedelta(days=days)
    limit_in_seconds = int(limit.strftime('%s'))
    for backup in _get_sorted_backups(backup_path):
        backup_timestamp = int(backup.stem)
        if backup_timestamp < limit_in_seconds:
            backup.unlink()
        else:
            break
