from nachos.common.exceptions import ConfigSyntaxException

import re


HOSTDEPENDENCY_PATTERN = re.compile(
    r'(host|hostgroup) "(.*?)" depends on (host|hostgroup) "(.*?)"'
    r'(?: and (host|hostgroup) "(.*?)")?'
)

SERVICEDEPENDENCY_PATTERN = re.compile(
    r'(host|hostgroup) service "(.*?);(.*?)" depends on (host|hostgroup) '
    r'service "(.*?);(.*?)"'
)


def get_config_key(change, changes, nacoma, mapper=None):
    object_type = change.object_type
    key_data = None

    if object_type == "contact_template":
        key_data = {"name": change.oldname}
        object_type = "contact"
    elif object_type == "host_template":
        key_data = {"name": change.oldname}
        object_type = "host"
    elif object_type == "service_template":
        key_data = {"name": change.oldname}
        object_type = "service"
    elif object_type == "command":
        key_data = {"command_name": change.oldname}
    elif object_type == "contact":
        key_data = {"contact_name": change.oldname}
    elif object_type == "contactgroup":
        key_data = {"contactgroup_name": change.oldname}
    elif object_type == "host":
        key_data = {"host_name": change.oldname}
    elif object_type == "hostgroup":
        key_data = {"hostgroup_name": change.oldname}
    elif object_type == "service":
        key_data = _get_service_config_key(
            nacoma, changes, change=change, mapper=mapper
        )
    elif object_type == "servicegroup":
        key_data = {"servicegroup_name": change.oldname}
    elif object_type == "timeperiod":
        key_data = {"timeperiod_name": change.oldname}
    elif object_type == "hostdependency":
        if change.oldname:
            key_data = _parse_hostdependency_name(change.oldname)
    elif object_type == "servicedependency":
        if change.oldname:
            key_data = _parse_servicedependency_name(change.oldname)
    elif object_type == "hostescalation":
        key_data = dict()
        host_id = change.get_from_oldstate("host_name")
        hostgroup_id = change.get_from_oldstate("hostgroup")

        # I dont' think hostgroup escalations are supported in monitor
        # but since there is such a field in Nacoma and it's supported
        # by Naemon I'll include it just in case.
        key_data = {
            "host_name": _get_host_name(
                host_id, changes, nacoma, mapper=(mapper.host if mapper else None)
            ),
            "hostgroup_name": _get_hostgroup_name(
                hostgroup_id,
                changes,
                nacoma,
                mapper=(mapper.hostgroup if mapper else None),
            ),
            "contacts": _get_contact_names(change, changes, nacoma, mapper=mapper),
            "contact_groups": _get_contactgroup_names(
                change, changes, nacoma, mapper=mapper
            ),
        }

    elif object_type == "serviceescalation":
        service_id = change.get_from_oldstate("service")

        key_data = {
            "contacts": _get_contact_names(change, changes, nacoma, mapper=mapper),
            "contact_groups": _get_contactgroup_names(
                change, changes, nacoma, mapper=mapper
            ),
        }

        key_data.update(
            _get_service_config_key(
                nacoma,
                changes,
                service_id=service_id,
                mapper=(mapper.service if mapper else None),
            )
        )

    if not key_data:
        raise ValueError(
            "Unable to get config_key for {} object with ID {}".format(
                object_type, change.object_id
            )
        )

    return (object_type, frozenset(key_data.items()))


def get_empty_key_data(object_type):
    if object_type in ("contact_template", "host_template", "service_template"):
        return {"name": None}
    if object_type == "command":
        return {"command_name": None}
    if object_type == "contact":
        return {"contact_name": None}
    if object_type == "contactgroup":
        return {"contactgroup_name": None}
    if object_type == "host":
        return {"host_name": None}
    if object_type == "hostgroup":
        return {"hostgroup_name": None}
    if object_type == "service":
        return {
            "service_description": None,
            "host_name": None,
            "hostgroup_name": None,
        }
    if object_type == "servicegroup":
        return {"servicegroup_name": None}
    if object_type == "timeperiod":
        return {"timeperiod_name": None}
    if object_type == "hostdependency":
        return {
            "host_name": None,
            "dependent_host_name": None,
            "hostgroup_name": None,
            "dependent_hostgroup_name": None,
        }
    if object_type == "servicedependency":
        return {
            "service_description": None,
            "host_name": None,
            "hostgroup_name": None,
            "dependent_service_description": None,
            "dependent_host_name": None,
            "dependent_hostgroup_name": None,
        }
    if object_type == "hostescalation":
        return {
            "host_name": None,
            "hostgroup_name": None,
            "contacts": None,
            "contact_groups": None,
        }
    if object_type == "serviceescalation":
        return {
            "service_description": None,
            "host_name": None,
            "hostgroup_name": None,
            "contacts": None,
            "contact_groups": None,
        }
    raise ConfigSyntaxException("Unexpected object type: {}".format(object_type))


def _get_contactgroup_names(change, changes, nacoma, mapper=None):
    if mapper:
        new_ids = change.get_newstate_members("contact_groups")
        old_ids = change.get_oldstate_members("contact_groups")
        contact_group_ids = set(
            cgroup.id
            for cgroup in mapper.contact_groups
            if not new_ids or cgroup.id not in new_ids
        )
        if old_ids:
            contact_group_ids.update(old_ids)
    else:
        # Let's hope oldstate has the info
        contact_group_ids = change.get_oldstate_members("contact_groups")
    if contact_group_ids:
        return ",".join(
            sorted(
                _get_contactgroup_name(cg_id, changes, nacoma)
                for cg_id in contact_group_ids
            )
        )
    return None


def _get_contact_names(change, changes, nacoma, mapper=None):
    if mapper:
        new_ids = change.get_newstate_members("contacts")
        old_ids = change.get_oldstate_members("contacts")
        contact_ids = set(
            contact.id
            for contact in mapper.contacts
            if not new_ids or contact.id not in new_ids
        )
        if old_ids:
            contact_ids.update(old_ids)
    else:
        # Let's hope oldstate has the info
        contact_ids = change.get_oldstate_members("contacts")
    if contact_ids:
        return ",".join(
            sorted(
                _get_contact_name(contact_id, changes, nacoma)
                for contact_id in contact_ids
            )
        )
    return None


def _get_service_config_key(nacoma, changes, service_id=None, change=None, mapper=None):

    description = None
    host_id = None
    host_name = None
    hostgroup_id = None
    hostgroup_name = None

    if not change and service_id:
        change = _get_object_from_changes("service", service_id, changes)

    if change:
        service_id = change.object_id
        _, description = change.oldname.split(";")
        hostgroup_id = change.get_from_oldstate("hostgroup_name")
        host_id = change.get_from_oldstate("host_name")

    if not host_id and not hostgroup_id:
        if not mapper and service_id:
            mapper = nacoma.service_by_id(service_id)

        if mapper:
            if not description:
                description = mapper.description
            if mapper.host:
                host_id = mapper.host.id
            if mapper.hostgroup:
                hostgroup_id = mapper.hostgroup.id

    if host_id:
        host_name = _get_host_name(host_id, changes, nacoma, mapper and mapper.host)

    if hostgroup_id:
        hostgroup_name = _get_hostgroup_name(
            hostgroup_id, changes, nacoma, mapper and mapper.hostgroup
        )

    if description and bool(host_name) != bool(hostgroup_name):
        return {
            "service_description": description,
            "host_name": host_name,
            "hostgroup_name": hostgroup_name,
        }

    return None


def _get_host_name(host_id, changes, nacoma, mapper=None):
    if not (host_id or mapper):
        return None
    change_host = _get_object_from_changes("host", host_id, changes)
    if change_host:
        return change_host.oldname
    nacoma_host = nacoma.host_by_id(host_id)
    if nacoma_host:
        return nacoma_host.host_name
    if mapper:
        return mapper.host_name
    raise ValueError("Unable to find host with ID {}".format(host_id))


def _get_contact_name(contact_id, changes, nacoma):
    if not contact_id:
        return None
    change_contact = _get_object_from_changes("contact", contact_id, changes)
    if change_contact:
        return change_contact.oldname
    nacoma_contact = nacoma.contact_by_id(contact_id)
    if nacoma_contact:
        return nacoma_contact.contact_name
    raise ValueError("Unable to find contact with ID {}".format(contact_id))


def _get_contactgroup_name(contactgroup_id, changes, nacoma):
    if not contactgroup_id:
        return None
    change_contactgroup = _get_object_from_changes(
        "contactgroup", contactgroup_id, changes
    )
    if change_contactgroup:
        return change_contactgroup.oldname
    nacoma_contactgroup = nacoma.contactgroup_by_id(contactgroup_id)
    if nacoma_contactgroup:
        return nacoma_contactgroup.contactgroup_name
    raise ValueError("Unable to find contactgroup with ID {}".format(contactgroup_id))


def _get_hostgroup_name(hostgroup_id, changes, nacoma, mapper=None):
    if not hostgroup_id:
        return None
    change_hostgroup = _get_object_from_changes("hostgroup", hostgroup_id, changes)
    if change_hostgroup:
        return change_hostgroup.oldname
    nacoma_hostgroup = nacoma.hostgroup_by_id(hostgroup_id)
    if nacoma_hostgroup:
        return nacoma_hostgroup.hostgroup_name
    if mapper:
        return mapper.hostgroup_name
    raise ValueError("Unable to find hostgroup with ID {}".format(hostgroup_id))


def _get_object_from_changes(object_type, object_id, changes):
    for change in changes:
        if change.object_type == object_type and change.object_id == object_id:
            return change


def _parse_hostdependency_name(name):
    # But define the pattern globally so we don't have to recompile it all the time.
    match = HOSTDEPENDENCY_PATTERN.match(name)
    if not match:
        raise NotImplementedError(
            "Badly formatted hostdependency name: {}".format(name)
        )
    (
        dependent_type,
        dependent_name,
        master_type_1,
        master_name_1,
        master_type_2,
        master_name_2,
    ) = match.groups()
    if master_type_1 == master_type_2:
        raise NotImplementedError(
            "Hostdependency on two objects of the same type not allowed: {}".format(
                name
            )
        )
    hostdependency = get_empty_key_data("hostdependency")
    if dependent_type == "host":
        hostdependency["dependent_host_name"] = dependent_name
    else:
        hostdependency["dependent_hostgroup_name"] = dependent_name
    if master_type_1 == "host":
        hostdependency["host_name"] = master_name_1
    else:
        hostdependency["hostgroup_name"] = master_name_1
    if master_type_2:
        if master_type_2 == "host":
            hostdependency["host_name"] = master_name_2
        else:
            hostdependency["hostgroup_name"] = master_name_2
    return hostdependency


def _parse_servicedependency_name(name):
    match = SERVICEDEPENDENCY_PATTERN.match(name)
    if not match:
        raise NotImplementedError(
            "Badly formatted servicedependency name: {}".format(name)
        )
    (
        dependent_parent_type,
        dependent_parent_name,
        dependent_service_description,
        master_parent_type,
        master_parent_name,
        master_service_description,
    ) = match.groups()
    servicedependency = get_empty_key_data("servicedependency")
    if dependent_parent_type == "host":
        servicedependency["dependent_host_name"] = dependent_parent_name
    else:
        servicedependency["dependent_hostgroup_name"] = dependent_parent_name
    if master_parent_type == "host":
        servicedependency["host_name"] = master_parent_name
    else:
        servicedependency["hostgroup_name"] = master_parent_name
    servicedependency["dependent_service_description"] = dependent_service_description
    servicedependency["service_description"] = master_service_description
    return servicedependency
