from collections import defaultdict
from enum import Enum


class ChangeStatus(Enum):
    DELETED = 1
    DONE = 2


class UpdateType(Enum):
    NEW = 1
    CHANGED = 2


class ChangedObjects(dict):
    """
    A custom dict for the changed objects in our `Updates` class. This will
    raise an error if we try to set a config_key that already exists, since
    that means we have two different objects with the same config_key. And
    that we can't support. The exception is when the value is
    `ChangeStatus.DONE`, which happens when we mark the config_key as
    processed.
    """

    def __setitem__(self, key, value):
        if value is not ChangeStatus.DONE and key in self:
            raise NotImplementedError(
                "Multiple {}s have identical key data".format(key[0])
            )
        dict.__setitem__(self, key, value)


class Updates:
    """
    Class for mapping changes to what config file they affect, and if they
    are new objects, updated objects or deleted objects.
    """

    def __init__(self):
        self._updates = defaultdict(
            lambda: {UpdateType.NEW: [], UpdateType.CHANGED: ChangedObjects()}
        )

    def new(self, filename, mapper):
        """
        Marks a new object to be added.

        :param filename: The config file it will be added to
        :param mapper: The database object to write
        """
        self._updates[filename][UpdateType.NEW].append(mapper)

    def changed(self, filename, config_key, mapper):
        """
        Marks a changed object for update.

        :param filename: The config file it is saved in
        :param config_key: The unique data identifying the object definition
        :param mapper: The database object to write
        """
        self._updates[filename][UpdateType.CHANGED][config_key] = mapper

    def deleted(self, filename, config_key):
        """
        Marks a deleted object for removal.

        :param filename: The config file it is saved in
        :param config_key: The unique data identifying the object definition
        """
        self._updates[filename][UpdateType.CHANGED][config_key] = ChangeStatus.DELETED

    def items(self):
        """
        Generates tuples of all files and their updates. The tuples yielded
        have a format of: `(filename, (new, changed)`

        Where `filename` is the config file, `new` is a list of all objects to
        add to that file, and `changed` is a dict-like ChangedObject instance
        with all changed objects in that file. The `changed` dict maps a
        `config_key` to either the database object to write or a
        `ChangeStatus` enum value.
        """
        for filename, changes in self._updates.items():
            yield (filename, (changes[UpdateType.NEW], changes[UpdateType.CHANGED]))
