from nachos.nacoma.mappers import BaseHostdependencyMapper
from nachos.engine.export.writers.helpers import (write_property,
                                                  write_bool,
                                                  write_subprop)


def write_hostdependency(hostdependency, out):
    """ Takes a mapped hostdependency object and writes it to a configuration
        file in a Naemon compatible format

        :param hostdependency: A mapped hostdependency object
        :param out: File handler
    """
    validate_hostdependency(hostdependency)

    if getattr(hostdependency, 'dependent_host', None) is not None:
        out.write("# hostdependency for host '{}'\n".format(
            hostdependency.dependent_host.host_name))
    elif getattr(hostdependency, 'dependent_hostgroup', None) is not None:  # noqa: E501
        out.write("# hostdependency for hostgroup '{}'\n".format(
            hostdependency.dependent_hostgroup.hostgroup_name))

    out.write('define hostdependency {\n')

    if getattr(hostdependency, 'dependent_host', None) is not None:
        write_subprop(out, 'dependent_host_name',
                      hostdependency.dependent_host, 'host_name')
    elif getattr(hostdependency, 'dependent_hostgroup', None) is not None:  # noqa: E501
        write_subprop(out, 'dependent_hostgroup_name',
                      hostdependency.dependent_hostgroup, 'hostgroup_name')

    write_subprop(out, 'host_name', hostdependency.host, 'host_name')
    write_subprop(out, 'hostgroup_name', hostdependency.hostgroup,
                  'hostgroup_name')
    write_bool(out, 'inherits_parent', hostdependency.inherits_parent)
    write_property(out, 'execution_failure_criteria',
                   hostdependency.execution_failure_criteria)
    write_property(out, 'notification_failure_criteria',
                   hostdependency.notification_failure_criteria)
    write_subprop(out, 'dependency_period', hostdependency.dependency_period,
                  'timeperiod_name')
    out.write('}\n')


def validate_hostdependency(hostdependency):
    """ Validates a hostdependency object

        A hostdependency can be on either a dependent host or a dependent
        hostgroup, so at least one need to exist and have a
        host/hostgroup name. It is not possible to have a hostdependency with
        both a dependent host and a dependent hostgroup.

        :param hostdependency: A mapped hostdependency object
    """

    assert isinstance(hostdependency, BaseHostdependencyMapper), \
        'Expected hostdependency parameter to implement ' \
        'BaseHostdependencyMapper'

    assert hostdependency.host or hostdependency.hostgroup, \
        'Hostdependency host and hostgroup cannot both be None'

    hostdependency_dependent_host = getattr(hostdependency,
                                            'dependent_host', None)
    hostdependency_dependent_hostgroup = getattr(hostdependency,
                                                 'dependent_hostgroup', None)

    assert hostdependency_dependent_host is not None or \
        hostdependency_dependent_hostgroup is not None, \
        'A hostdependency needs either a dependent host or dependent hostgroup'

    if hostdependency_dependent_host is not None:
        assert getattr(hostdependency.dependent_host, 'host_name', None) \
            is not None, 'Hostdependency dependent host name cannot be None'

    if hostdependency_dependent_hostgroup is not None:
        assert getattr(hostdependency.dependent_hostgroup,
                       'hostgroup_name', None) is not None, \
            'Hostdependency dependent hostgroup name cannot be None'
