from nachos.nacoma.mappers import BaseServicedependencyMapper
from nachos.engine.export.writers.helpers import (write_property,
                                                  write_bool,
                                                  write_subprop)


def write_servicedependency(servicedependency, out):
    """ Takes a mapped servicedependency object and writes it to a
        configuration file in a Naemon compatible format

        :param servicedependency: A mapped servicedependency object
        :param out: File handler
    """
    validate_servicedependency(servicedependency)

    if getattr(servicedependency, 'dependent_host', None) is not None:
        out.write("# servicedependency for service '{}' on host '{}'\n".format(
            servicedependency.dependent_service.description,
            servicedependency.dependent_host.host_name))

    if getattr(servicedependency, 'dependent_hostgroup', None) is not None:
        out.write("# servicedependency for service '{}' on hostgroup '{}'\n"
                  .format(servicedependency.dependent_service.description,
                          servicedependency.dependent_hostgroup.hostgroup_name)
                  )

    out.write('define servicedependency {\n')

    if getattr(servicedependency, 'dependent_host', None) is not None:
        write_subprop(out, 'dependent_host_name',
                      servicedependency.dependent_host, 'host_name')
    if getattr(servicedependency, 'dependent_hostgroup', None) is not None:
        write_subprop(out, 'dependent_hostgroup_name',
                      servicedependency.dependent_hostgroup,
                      'hostgroup_name')

    write_subprop(out, 'dependent_service_description',
                  servicedependency.dependent_service,
                  'description')

    if getattr(servicedependency, 'host', None) is not None:
        write_subprop(out, 'host_name', servicedependency.host, 'host_name')
    if getattr(servicedependency, 'hostgroup', None) is not None:
        write_subprop(out, 'hostgroup_name', servicedependency.hostgroup,
                      'hostgroup_name')

    write_subprop(out, 'service_description',
                  servicedependency.service,
                  'description')
    write_bool(out, 'inherits_parent', servicedependency.inherits_parent)
    write_property(out, 'execution_failure_criteria',
                   servicedependency.execution_failure_criteria)
    write_property(out, 'notification_failure_criteria',
                   servicedependency.notification_failure_criteria)
    write_subprop(out, 'dependency_period',
                  servicedependency.dependency_period, 'timeperiod_name')
    out.write('}\n')


def validate_servicedependency(servicedependency):
    """ Validates a servicedependency object

        A servicedependency can be on either a host or a hostgroup, so at least
        one need to exist and have a host/hostgroup name. It can have either a
        dependent host or dependent hostgroup. It is not possible
        to have a servicedependency with both a host and a hostgroup, or both
        dependent host and dependent hostgroup.

        :param servicedependency: A mapped servicedependency object
    """

    assert isinstance(servicedependency, BaseServicedependencyMapper), \
        'Expected servicedependency parameter to implement ' \
        'BaseServicedependencyMapper'

    assert getattr(servicedependency.service,
                   'description', None) is not None, \
        'Servicedependency service_description cannot be None'

    assert getattr(servicedependency.dependent_service,
                   'description', None) is not None, \
        'Servicedependency dependent_service_description cannot be None'

    servicedependency_host = getattr(servicedependency, 'host', None)
    servicedependency_hostgroup = getattr(servicedependency,
                                          'hostgroup', None)
    servicedependency_dependent_host = getattr(
        servicedependency,
        'dependent_host', None)
    servicedependency_dependent_hostgroup = getattr(
        servicedependency,
        'dependent_hostgroup', None)

    assert servicedependency_host is not None or \
        servicedependency_hostgroup is not None, \
        'A servicedependency needs either a host or hostgroup'

    assert servicedependency_dependent_host is not None or \
        servicedependency_dependent_hostgroup is not None, \
        'A servicedependency needs either a dependent host ' \
        'or dependent hostgroup'

    if servicedependency_host is not None:
        assert getattr(servicedependency.host, 'host_name', None), \
            'servicedependency host name cannot be None'

    if servicedependency_hostgroup is not None:
        assert getattr(servicedependency.hostgroup, 'hostgroup_name', None), \
            'servicedependency hostgroup name cannot be None'

    if servicedependency_dependent_host is not None:
        assert getattr(servicedependency.dependent_host, 'host_name', None), \
            'servicedependency dependent host name cannot be None'

    if servicedependency_dependent_hostgroup is not None:
        assert getattr(servicedependency.dependent_hostgroup, 'hostgroup_name',
                       None), \
            'servicedependency dependent hostgroup name cannot be None'
