import datetime

from oslo_log import log
from oslo_config import cfg
from phpserialize import dumps, loads
from sqlalchemy import inspect
from sqlalchemy.orm import joinedload
from sqlalchemy.orm.interfaces import MANYTOMANY, MANYTOONE

from nachos.nacoma import Base
from nachos.nacoma.api.sqlalchemy.models import (ActionLog,
                                                 Changelog,
                                                 ChangelogHistory,
                                                 Command,
                                                 Contactgroup,
                                                 Contact,
                                                 Customvar,
                                                 File,
                                                 Host,
                                                 Hostdependency,
                                                 Hostescalation,
                                                 Hostgroup,
                                                 Objlock,
                                                 Service,
                                                 Servicedependency,
                                                 Serviceescalation,
                                                 Servicegroup,
                                                 Timeperiod)
from nachos.nacoma.common.utils import (
    get_nacoma_class,
    get_nacoma_mapper,
)
from nachos.nacoma.mappers import (ChangelogMapper,
                                   CommandMapper,
                                   ContactMapper,
                                   ContactgroupMapper,
                                   CustomvarMapper,
                                   FileMapper,
                                   HostMapper,
                                   HostdependencyMapper,
                                   HostescalationMapper,
                                   HostgroupMapper,
                                   ServiceMapper,
                                   ServicedependencyMapper,
                                   ServiceescalationMapper,
                                   ServicegroupMapper,
                                   TimeperiodMapper)
from nachos.nacoma.exceptions.exceptions import NacomaLockAquireFail


logger = log.getLogger(__name__)


class SQLAlchemyNacoma(Base):
    _session = None

    def set_session(self, session):
        """ Inject the session dependency used to access the database

        :param session: SQLAlchemy session to the Nacoma database """
        self._session = session
        self._session.autoflush = False
        self._session.autocommit = False

    def aquire_database_lock(self, object_type='all', username=None):  # noqa: E501
        """ Inserts a record into the objlocks table

        :param object_type: Nacoma legacy - The only type Nacoma is aware of
                         (and locks the database) is the 'all' type.
        :param username: The user who triggered the export """

        assert username is not None

        # Check if the database is already locked
        lock = self._session.query(Objlock) \
            .filter(Objlock.object_type == object_type) \
            .order_by(Objlock.time.desc()).first()

        if lock is not None and datetime.datetime.now() < lock.time:
            raise NacomaLockAquireFail('Lock is already aquired')

        # Create a lock, time given is the time when the lock expires
        logger.debug('Locking Nacoma database')
        new_lock = Objlock(object_type=object_type,
                           object_id=0,
                           username=username,
                           time=(datetime.datetime.now() +
                                 datetime.timedelta(minutes=10)))
        self._session.add(new_lock)
        self._session.commit()

        # Make sure we got the latest lock in case of race conditions
        newest_lock = self._session.query(Objlock) \
            .filter(Objlock.object_type == object_type) \
            .order_by(Objlock.time.desc()).first()

        if newest_lock is None:
            raise NacomaLockAquireFail('Failed to aquire lock')
        if newest_lock.id != new_lock.id:
            raise NacomaLockAquireFail('Lock is already aquired')

    def release_database_lock(self, object_type='all'):  # noqa: E501
        """ Delete the records in the objlocks table for a specific type

        :param object_type: Nacoma legacy - The only type Nacoma is aware of
                     (and locks the database) is the 'all' type. """
        logger.debug('Unlocking Nacoma database')
        self._session.query(Objlock).filter(
            Objlock.object_type == object_type).delete()
        self._session.commit()

    def archive_changes(self, username=None):
        """
        Archives entries in the change_log table and commits them to
        changelog_history
        :param username: The username for the user whose changes needs to be
        archived
        :return:
        """
        logger.debug('Archiving changelog')
        full_changelog = self._fetch_changelog_entries()
        if username:
            changelog = self._fetch_changelog_entries(username)
        else:
            changelog = full_changelog
        if not changelog:
            return

        for entry in changelog:
            changelog_history = self._generate_history_from_changelog(
                entry, full_changelog
            )
            self._session.add(changelog_history)
            self._session.delete(entry)
        self._session.commit()

    def log_action(self, action_type, username):
        """ Creates an action_log entry in the Nacoma database. If the
            configuration file timestamp is newer than the timestamp of the
            latest action_log entry, Nacoma will try to reimport the
            configuration.

        :param action_type: String import/export - Whether configuration is
                            imported or exported
        :param username: String Username of the user who executed the action
        """
        assert username is not None
        # We can't have an enum called 'import' because it's a keyword.
        # This could probably be done in a nicer way.
        assert action_type == 'export' or action_type == 'import'

        log = ActionLog(username=username, action=action_type)
        self._session.add(log)
        self._session.commit()
        logger.debug('Logged Nacoma action "{}" by {}'
                     .format(action_type, username))

    def apply_changelog(self, username=None):
        """ Applies the changelog entries to the database.

        Changelog entries contain a newstate -- a serialized associative array
        containing all the new values for a Nacoma object. These are used to
        update objects to the new state.

        :param username: String Only apply changelog entries belonging to the
                         specified username. If None all changelog entries are
                         applied.
        """
        changes = self._fetch_changelog_entries(username)
        changes.sort(reverse=False, key=lambda c: c.id)
        logger.debug('Applying changelog for {}'
                     .format(username if username is not None else 'everyone'))
        for entry in changes:
            try:
                logger.debug('Apply changelog {} for {} (id: {}): {} => {}'
                             .format(entry.id,
                                     entry.object_type,
                                     entry.object_id,
                                     entry.oldstate,
                                     entry.newstate))
                self._transition_object_to_state(entry.object_type,
                                                 entry.object_id,
                                                 entry.oldstate,
                                                 entry.newstate)
            except (LookupError, ValueError) as e:
                logger.warning(e)
                continue
        self._session.flush()

    def rollback(self):
        """ Rollbacks the current transaction in progress.
        """
        self._session.rollback()

    def rollback_changelog(self):
        """ Rollbacks the changelog entries

        Changelog entries contain an oldstate -- a serialized associative array
        containing all the old values for a Nacoma object. These are used to
        revert objects to their old state.
        """

        """
            If a host is removed in nacoma the belonging services will be
            removed after the host is removed. This means that the changelog
            will have records of the services being removed after the host.
            If we want to rollback a removal of the host, that is to recreate
            it, the services will get recreated first if we go by descending
            order. This will raise an error since we will try to assign the
            service to a host that does not yet exist. So we must reverse the
            order of changelog entries that will recreate objects
        """
        changes = self._fetch_changelog_entries()
        changes.sort(reverse=True, key=lambda c: c.id)
        other_changes = [c for c in changes if c.newstate is not None]
        recreate_changes = [c for c in changes if c.newstate is None]
        recreate_index = [i for i, c in enumerate(changes)
                          if c.newstate is None]

        # Reinsert the recreate changes in reverse order
        recreate_changes.reverse()
        [other_changes.insert(z[0], z[1]) for
         z in zip(recreate_index, recreate_changes)]
        changes = other_changes

        for entry in changes:
            try:
                logger.debug('Roll back changelog {} for {} (id: {}): {} => {}'
                             .format(entry.id,
                                     entry.object_type,
                                     entry.object_id,
                                     entry.newstate,
                                     entry.oldstate))
                self._transition_object_to_state(entry.object_type,
                                                 entry.object_id,
                                                 entry.newstate,
                                                 entry.oldstate)
            except (LookupError, ValueError) as e:
                logger.warning(e)
                continue
        self._session.flush()

    @property
    def changelogs(self):
        return self._all(Changelog, mapper=ChangelogMapper)

    def changelogs_by_username(self, username):
        return (
            ChangelogMapper(obj)
            for obj in self._session.query(Changelog)
            .filter_by(username=username)
            .all()
        )

    def changelog_by_id(self, changelog_id, eager=False):
        return self._first(Changelog, changelog_id, mapper=ChangelogMapper, eager=eager)

    @property
    def commands(self):
        return self._all(Command, mapper=CommandMapper, eager=True)

    def command_by_id(self, command_id, eager=False):
        return self._first(Command, command_id, mapper=CommandMapper, eager=eager)

    @property
    def contactgroups(self):
        return self._all(Contactgroup, mapper=ContactgroupMapper, eager=True)

    def contactgroup_by_id(self, contactgroup_id, eager=False):
        return self._first(
            Contactgroup, contactgroup_id, mapper=ContactgroupMapper, eager=eager
        )

    @property
    def contacts(self):
        return self._all(Contact, mapper=ContactMapper, eager=True)

    def contact_by_id(self, contact_id, eager=False):
        return self._first(Contact, contact_id, mapper=ContactMapper, eager=eager)

    @property
    def customvars(self):
        return self._all(Customvar, mapper=CustomvarMapper)

    @property
    def hostdependencies(self):
        return self._all(Hostdependency, mapper=HostdependencyMapper, eager=True)

    @property
    def hostescalations(self):
        return self._all(Hostescalation, mapper=HostescalationMapper, eager=True)

    @property
    def hostgroups(self):
        return self._all(Hostgroup, mapper=HostgroupMapper, eager=True)

    def hostgroup_by_id(self, hostgroup_id, eager=False):
        return self._first(Hostgroup, hostgroup_id, mapper=HostgroupMapper, eager=eager)

    @property
    def hosts(self):
        return self._all(Host, mapper=HostMapper, eager=True)

    def host_by_id(self, host_id, eager=False):
        return self._first(Host, host_id, mapper=HostMapper, eager=eager)

    @property
    def servicedependencies(self):
        return self._all(Servicedependency, mapper=ServicedependencyMapper, eager=True)

    @property
    def serviceescalations(self):
        return self._all(Serviceescalation, mapper=ServiceescalationMapper, eager=True)

    @property
    def servicegroups(self):
        return self._all(Servicegroup, mapper=ServicegroupMapper, eager=True)

    def servicegroup_by_id(self, servicegroup_id, eager=False):
        return self._first(
            Servicegroup, servicegroup_id, mapper=ServicegroupMapper, eager=eager
        )

    @property
    def services(self):
        return self._all(Service, mapper=ServiceMapper, eager=True)

    def service_by_id(self, service_id, eager=False):
        return self._first(Service, service_id, mapper=ServiceMapper, eager=eager)

    @property
    def timeperiods(self):
        return self._all(Timeperiod, mapper=TimeperiodMapper, eager=True)

    @property
    def files(self):
        return self._all(File, mapper=FileMapper)

    def file_by_id(self, file_id):
        return self._first(File, file_id, mapper=FileMapper)

    def object_by_type_and_id(self, object_name, object_id, eager=False):
        mapper = get_nacoma_mapper(object_name)
        object_type = get_nacoma_class(object_name)
        if mapper is None or object_type is None:
            return None
        return self._first(object_type, object_id, mapper=mapper, eager=eager)

    def _all(self, model, mapper=None, eager=False):
        query = self._query(model, eager=eager)
        if mapper:
            return (mapper(obj) for obj in query.all())
        return query.all()

    def _first(self, model, obj_id, mapper=None, eager=False):
        obj = self._query(model, eager=eager).filter_by(id=obj_id).first()
        return mapper(obj) if mapper and obj else obj

    def _query(self, model, eager=False):
        query = self._session.query(model)
        if eager and hasattr(model, "exported_relationships"):
            for join in model.exported_relationships():
                query = query.options(joinedload(join))
        return query

    def _transition_object_to_state(self, object_type, object_id,
                                    from_state, to_state):
        """ Transitions a Naemon object from one state to another using the
            old/newstate from changelog entries. The object can go from
            oldstate->newstate or the other way newstate->oldstate

        :param object_type: String the type of Naemon object
        :param object_id: Int object ID
        :param from_state: Dict with the state the object is coming from.
        :param to_state: Dict with the state the object is going to.
        """
        type = get_nacoma_class(object_type)
        if type is None:
            raise ValueError('Unknown object type "{}"'.format(object_type))
        if from_state is not None:
            from_state = loads(
                from_state, charset=cfg.CONF.nacoma.blob_encoding, decode_strings=True
            )
        if to_state is not None:
            to_state = loads(
                to_state, charset=cfg.CONF.nacoma.blob_encoding, decode_strings=True
            )
        if self._is_custom_var(type, from_state, to_state):
            """
                Customvars behave differently from regular changelog entries.
                It is up to the user to name the customvar key, so it can be
                named pretty much anything. The old/newstates also behave
                differently compared to regular changelog entries. In regular
                changelog entries, if you create a new object oldstate is None,
                and if you delete an object newstate is None. With customvars
                the states are only None if no prior customvars for that
                object exists. If prior customvars exists old/newstate will
                not be None, but the key will contain an empty value. If you
                in Nacoma set an empty value for a customvar it will be
                considered a deletion.
            """
            # In case any state is empty, copy keys (ONLY) from the other state
            if from_state is None and to_state is not None:
                from_state = dict.fromkeys(to_state)
            elif to_state is None and from_state is not None:
                to_state = dict.fromkeys(from_state)
            elif from_state is None and to_state is None:
                raise LookupError('No from_state nor to_state found')

            for key in sorted(from_state.keys()):
                assert key in to_state.keys()
                cvar = (self._session.query(Customvar)
                        .filter(Customvar.id == object_id,
                                Customvar.type == object_type,
                                Customvar.key == key)
                        .one_or_none())
                if not from_state[key] and to_state[key]:
                    logger.debug('Creating custom var {}={} on {} (id={})'.
                                 format(key, to_state[key],
                                        object_type, object_id))
                    # For some reason, if you delete an object with a
                    # custom var in Nacoma, the custom var is not removed
                    # from the DB. So instead we'll just update it here.
                    if cvar is None:
                        cvar = Customvar(type=object_type,
                                         id=object_id,
                                         key=key)
                    f = None
                    t = {'variable': key, 'value': to_state[key]}
                elif from_state[key] and not to_state[key]:
                    logger.debug('Deleting custom var {}={} on {} (id={})'.
                                 format(key, from_state[key],
                                        object_type, object_id))
                    if cvar is None:
                        # Due to a bug in Nacoma, the custom var may already
                        # have been removed. So if we don't find it for
                        # deletion, just continue
                        continue
                    f = {'variable': key, 'value': from_state[key]}
                    t = None
                elif from_state[key] and to_state[key]:
                    logger.debug('Transitioning custom var {}={} on {} '
                                 '(ID: {}): {} => {}'
                                 .format(key, from_state[key], object_type,
                                         object_id, from_state[key],
                                         to_state[key]))
                    if cvar is None:
                        raise LookupError('No custom var {} on {} (ID: {})'
                                          .format(key, object_type, object_id))
                    f = {'variable': key, 'value': from_state[key]}
                    t = {'variable': key, 'value': to_state[key]}
                self._changelog_entry_transition(object_id, cvar, f, t)
        else:
            object = (self._session.query(type)
                      .filter(type.id == object_id).one_or_none())
            if object is None:
                if from_state is None:
                    logger.debug('{} (id: {}) does not exist, create it.'
                                 .format(object_type, object_id))
                    object = type(id=object_id)
                else:
                    raise LookupError('Could not find {} ({}) for transition'
                                      .format(object_type, object_id))
            self._changelog_entry_transition(object_id, object,
                                             from_state, to_state)

    def _changelog_entry_transition(self, object_id, object,
                                    from_state, to_state):
        """
        Transition an object from one state to another.

        :param object: A SQLAlchemy model that should be transitioned.
        :param from_state: Dict with the state the object is coming from.
        :param to_state: Dict with the state the object is going to.
        :return:
        """
        if to_state is None:
            logger.debug('Changelog transition: Deleting object from DB {}'
                         .format(object))
            self._session.delete(object)
            return
        logger.debug('Changelog transition: Transitioning object {}'
                     .format(object))
        self._transition_to_state(object_id, object, from_state, to_state)
        if from_state is None:
            logger.debug('Changelog transition: Adding object to DB {}, id {}'
                         .format(object, object_id))
            self._session.add(object)
            self._session.flush()

    def _transition_to_state(self, object_id, object, from_state, to_state):  # noqa: E501
        """
        Transition an object from one state to another.

        :param object_id: Int object ID
        :param object: A SQLAlchemy model that should be transitioned.
        :param from_state: Dict with the state the object is coming from.
        :param to_state: Dict with the state the object is going to.
        :return:
        """
        inspection = inspect(type(object))

        for key, value in sorted(to_state.items()):
            assert from_state is None or key in from_state, \
                'Expected key "{}" to exist in from_state'.format(key)
            # We need to be able to map column name to attributes.
            # Therefore models that have attributes not matching the column
            # name needs synonyms. To avoid conflicts of any kind, we
            # use a _ as a prefix and a _syn for suffix for all synonyms.
            syn = getattr(inspection.synonyms, '_' + key + '_syn', None)
            attr = syn.name if syn is not None else key

            assert hasattr(object, attr), 'No such attribute "{}"'.format(attr)

            # Check if the new value should be an ORM object or just a value
            relationship = getattr(inspection.relationships, attr, None)
            if relationship is None:
                # No relationship, but make sure the new value is
                # correctly formatted.

                # Detect if it is an options field
                if isinstance(value, dict) or from_state is not None and \
                        isinstance(from_state[key], dict):

                    curr_val = getattr(object, attr)
                    curr_set = set(curr_val.split(',')) \
                        if isinstance(curr_val, str) and len(curr_val) > 0 \
                        else set()
                    from_set = set(from_state[key].values()) \
                        if from_state is not None and \
                        isinstance(from_state[key], dict) else set()
                    to_set = set(value.values()) \
                        if isinstance(value, dict) else set()
                    new_set = (curr_set | to_set) - from_set
                    value = ','.join(sorted(new_set))

                if isinstance(value, str) and len(value) == 0:
                    value = None

                # Nacoma sometimes doesn't store values as proper type in
                # changelog state array, so make sure we cast it properly by
                # checking the column value type in the model definition.
                if value is not None:
                    attr_type = getattr(inspection.c, attr).type.python_type
                    if attr_type == bytes:
                        value = attr_type(value.encode(cfg.CONF.nacoma.blob_encoding))
                    elif attr_type == bool:
                        if isinstance(value, str) and value.isnumeric():
                            value = int(value)
                        value = bool(value)
                    else:
                        # Cloned objects might have the same id as the original
                        # object in the newstate. This should not overwrite the
                        # object_id
                        if key == 'id':
                            continue
                        value = attr_type(value)
            else:
                # The value should be one of multiple orm objects, find them
                mapper = relationship.mapper
                id = mapper.all_orm_descriptors.id
                if relationship.direction == MANYTOMANY:
                    # The old and new state in the Nacoma changelog table
                    # only shows the delta between the two states.
                    # Meaning that if for instance a host has 2 children and
                    # 1 is removed. The oldstate will only show the child that
                    # was removed and show no children in the newstate.
                    # So here we calculate the union of the delta and the
                    # current relationships.

                    # To be able to query with sqlalchemy the id's of related
                    # object(s) must be contained within a list
                    if isinstance(from_state, dict):
                        from_values = from_state.get(key, [])
                        if isinstance(from_values, dict):
                            from_values = from_values.values()
                        elif from_values == '':
                            from_values = []
                    else:
                        from_values = []

                    if value:
                        to_values = value.values()
                    else:
                        to_values = []

                    # Apply the delta to the current relationships
                    current_set = set(getattr(object, attr))
                    from_set = set(self._session.query(mapper)
                                   .filter(id.in_(from_values)).all())
                    f0 = set([int(v) for v in from_values])
                    f1 = set([int(o.id) for o in from_set])
                    if f0 != f1:
                        logger.warning('Cannot create expected relationship. '
                                       'Wanted {} with Ids {} but got Ids {}'
                                       .format(mapper.class_.__name__, f0, f1))
                    to_set = set(self._session.query(mapper)
                                 .filter(id.in_(to_values)).all())
                    t0 = set([int(v) for v in to_values])
                    t1 = set([int(o.id) for o in to_set])
                    if t0 != t1:
                        logger.warning('Cannot create expected relationship. '
                                       'Wanted {} with Ids {} but got Ids {}'
                                       .format(mapper.class_.__name__, t0, t1))
                    value = list((current_set | to_set) - from_set)

                elif relationship.direction == MANYTOONE:
                    if value:
                        val = self._session.query(mapper) \
                            .filter(id == value).one_or_none()
                        if val is None:
                            self._session.rollback()
                            raise LookupError('Could not create relationship '
                                              'to {} with id {}'
                                              .format(mapper.class_.__name__,
                                                      value))
                        value = val
                    else:
                        if isinstance(object, Hostdependency):
                            """
                            Hostdependencies are special in Nacoma. Host and
                            Hostgroup are part of a compound key, so they
                            can not be None. But in practice you only need
                            either Host or Hostgroup, so nacoma just sets
                            Host/Hostgroup id to 0 if one of them is empty.
                            """
                            if relationship.mapper.class_ == Host:
                                logger.debug("Empty hostdependency host")
                                attr = '_host_fk'
                                value = 0
                            elif relationship.mapper.class_ == Hostgroup:
                                logger.debug("Empty hostdependency hostgroup")
                                attr = '_hostgroup_fk'
                                value = 0
                        else:
                            value = None
            logger.debug('Setting attribute "{}" = {}'.format(attr, value))
            setattr(object, attr, value)

    def _fetch_changelog_entries(self, username=None):
        """
        Given a username, fetch all changelog entries related to this user.
        Some entries may be done by other users, but objects that have been
        altered by multiple users, this method returns changelog entries by
        those users as well.
        If no username is given, this method simply returns all changelog
        entries in an appropriate order.

        :param username: The username to fetch changelog entries for
        :return List(Changelog)
        """
        q = self._session.query(Changelog)
        if username is None:
            return q.all()
        changed_objects = []
        for change in q.filter(Changelog.username == username).all():  # noqa: E711 E501
            identifier = (change.object_type, change.object_id)
            if identifier not in changed_objects:
                changed_objects.append(identifier)
        changelog = []
        for change in q.all():
            identifier = (change.object_type, change.object_id)
            if identifier in changed_objects:
                changelog.append(change)
        return changelog

    def _is_custom_var(self, object_type, from_state, to_state):  # noqa: E501
        """ Try to identify if it is a custom variable or not

            Custom variables start with _ and is UPPERCASE, except timeperiod
            exclusions. If the object_type is Timeperiod and the attribute does
            not exist for the object_type, and there is no synonym for it, we
            assume it is a timeperiod exclude. I.e. customvar.

            :param object_type: Nacoma SQLAlchemy model
            :param state: A deserialized oldstate/newstate from a Nacoma
                          changelog entry

            :return: True if customvar, else False
        """
        if from_state is None and to_state is not None:
            state = to_state
        elif from_state is not None and to_state is None:
            state = from_state
        elif from_state is not None and to_state is not None:
            state = from_state
        else:
            return False
        for key in state.keys():
            if key.startswith('_') and key.isupper():
                return True
            if (object_type == Timeperiod
                    and not hasattr(object_type, key)
                    and not hasattr(object_type, "_{}_syn".format(key))):
                return True
        return False

    def _generate_history_from_changelog(self, changelog_entry, full_changelog):
        """ Generates a changelog_history entry based on a changelog entry.

        :param changelog_entry: SQLAlchemy Changelog object
        :return: SQLAlchemy ChangelogHistory object
        """
        assert isinstance(changelog_entry, Changelog)
        return ChangelogHistory(
            object_type=changelog_entry.object_type,
            object_name=changelog_entry.newname,
            oldname=changelog_entry.oldname,
            newname=changelog_entry.newname,
            oldstate=self._to_historic_state(
                changelog_entry.oldstate, changelog_entry.object_type, full_changelog
            ),
            newstate=self._to_historic_state(
                changelog_entry.newstate, changelog_entry.object_type, full_changelog
            ),
            username=changelog_entry.username,
            time=changelog_entry.time
        )

    def _to_historic_state(self, state, object_type, full_changelog):
        """ Change all numeric ID references used in a state blob to names instead.

        :param state: bytes The state blob to update with names
        :param object_type: str The object type the state belongs to
        :param full_changelog: list The changelog used to get names of deleted objects
        :return: bytes The (possibly) updated state blob
        """

        # Decide on what state fields we need to look at for this object.
        if object_type == "host":
            fields = {
                "parents": "host",
                "children": "host",
                "contacts": "contact",
                "contact_groups": "contactgroup",
                "hostgroups": "hostgroup",
            }
        elif object_type == "service":
            fields = {
                "parents": "service",
                "children": "service",
                "contacts": "contact",
                "contact_groups": "contactgroup",
                "servicegroups": "servicegroup",
            }
        elif object_type == "hostgroup":
            fields = {
                "members": "host",
                "hostgroup_members": "hostgroup",
            }
        elif object_type == "servicegroup":
            fields = {
                "members": "service",
                "servicegroup_members": "servicegroup",
            }
        elif object_type == "contact":
            fields = {
                "contactgroups": "contactgroup",
            }
        elif object_type == "contactgroup":
            fields = {
                "members": "contact",
                "contactgroup_members": "contactgroup",
            }
        else:
            # Other object types require no state change.
            return state

        # If the state can't be decoded it requires no change (it's probably empty).
        try:
            decoded_state = loads(
                state, charset=cfg.CONF.nacoma.blob_encoding, decode_strings=True
            )
        except ValueError:
            return state

        # Check if our state contains any of the member fields we need to change.
        for field, field_type in fields.items():
            field_state = decoded_state.get(field)
            if type(field_state) is dict:
                # Change each numeric ID value to the corresponding object name.
                for index, object_id in field_state.items():
                    try:
                        object_id = int(object_id)
                    except ValueError:
                        # If the ID doesn't look like an integer we'll leave it as is.
                        continue
                    # Usually we should be able to find the name in the database.
                    obj = self.object_by_type_and_id(field_type, object_id)
                    if obj:
                        obj_name = obj.changelog_name
                    else:
                        # But if the object was deleted after this particular change,
                        # then we won't find the object in the database. In that case
                        # it should be possible to get the name from the changelog
                        # entry for the delete operation instead.
                        change = self._find_change_object(
                            field_type, object_id, full_changelog
                        )
                        obj_name = change and change.oldname
                    # Update the state value with the name, if we found one.
                    if obj_name:
                        field_state[index] = obj_name

        # Re-encode the (possibly) updated state and return it.
        return dumps(decoded_state, charset=cfg.CONF.nacoma.blob_encoding)

    def _find_change_object(self, object_type, object_id, changes):
        """ Find the first matching change item for a given object.

        :param object_type: str The object type of the object we're looking for
        :param object_id: int The object ID of the object we're looking for
        :param changes: list The changelog to look through
        :return: ChangelogMapper The first changelog item matching `object_type` and
            `object_id` in the list of `changes`, otherwise None
        """
        for change in changes:
            if change.object_type == object_type and change.object_id == object_id:
                return change
        return None
