"""
SQLAlchemy models for the nacoma's database tables
"""
from sqlalchemy import (BLOB, Boolean, Column, DateTime, Float, ForeignKey,
                        Integer, LargeBinary,  SmallInteger, String, Text)
from sqlalchemy.orm import relationship, synonym
from sqlalchemy.sql import func

from nachos.nacoma.api.sqlalchemy.associations import (contact2contactgroup,
                                                       contactgroup2member,
                                                       host2contact,
                                                       host2contactgroup,
                                                       host2hostgroup,
                                                       host2parent,
                                                       hostescalation2contact,
                                                       hostescalation2contactgroup,  # noqa: E501
                                                       hostgroup2member,
                                                       service2contact,
                                                       service2contactgroup,
                                                       service2servicegroup,
                                                       serviceescalation2contact,  # noqa: E501
                                                       serviceescalation2contactgroup,  # noqa: E501
                                                       servicegroup2member,
                                                       timeperiod2exclude)
from nachos.nacoma.api.sqlalchemy.base import Base
from nachos.db.sqlalchemy.types import Epoch


class ActionLog(Base):
    __tablename__ = "action_log"

    username = Column(String(length=30), primary_key=True)
    action = Column(String(length=50), primary_key=True)
    time = Column(DateTime, server_default=func.now(), primary_key=True)


class Changelog(Base):
    __tablename__ = "changelog"

    id = Column(Integer, primary_key=True)

    object_type = Column(String(length=100))
    object_id = Column(Integer)

    oldname = Column(String(length=255))
    newname = Column(String(length=255))

    oldstate = Column(BLOB)
    newstate = Column(BLOB)

    username = Column(String(length=255))
    time = Column(DateTime, server_default=func.now())


class ChangelogHistory(Base):

    __tablename__ = "changelog_history"

    id = Column(Integer, primary_key=True)

    object_type = Column(String(length=100))
    object_name = Column(String(length=255))

    oldname = Column(String(length=255))
    newname = Column(String(length=255))

    oldstate = Column(BLOB)
    newstate = Column(BLOB)

    username = Column(String(length=255))
    time = Column(DateTime, server_default=func.now())


class File(Base):
    __tablename__ = "file_tbl"

    id = Column(Integer, primary_key=True)
    name = Column('file_name', String(length=255))

    _file_name_syn = synonym('file_name')


class Command(Base):
    __tablename__ = "command"

    id = Column(Integer, primary_key=True)

    command_line = Column(LargeBinary)
    file = relationship(File, lazy='select')
    command_name = Column(String(length=255))
    register = Column(Boolean)

    _file_fk = Column('file_id', Integer, ForeignKey(File.id))

    _file_id_syn = synonym('file')

    @classmethod
    def exported_relationships(cls):
        return (cls.file,)


class Timeperiod(Base):
    __tablename__ = "timeperiod"

    id = Column(Integer, primary_key=True)

    alias = Column(String(length=255), default='')
    excluded = relationship('Timeperiod', secondary=timeperiod2exclude,
                            primaryjoin=id == timeperiod2exclude.c.timeperiod,
                            secondaryjoin=id == timeperiod2exclude.c.exclude,
                            lazy='select')
    custom_vars = \
        relationship('Customvar',
                     primaryjoin="and_(Timeperiod.id==Customvar.id,"
                                 "Customvar.type=='timeperiod')",
                     foreign_keys='[Customvar.id, Customvar.type]',
                     cascade='all, delete-orphan',
                     lazy='select',
                     overlaps='custom_vars')
    file = relationship(File, lazy='select')
    timeperiod_name = Column(String(length=255))
    register = Column(Boolean)

    monday = Column(String(length=255))
    tuesday = Column(String(length=255))
    wednesday = Column(String(length=255))
    thursday = Column(String(length=255))
    friday = Column(String(length=255))
    saturday = Column(String(length=255))
    sunday = Column(String(length=255))

    _file_fk = Column('file_id', Integer, ForeignKey(File.id))

    _file_id_syn = synonym('file')
    _exclude_syn = synonym('excluded')

    @classmethod
    def exported_relationships(cls):
        return (cls.custom_vars, cls.excluded, cls.file)


class Contact(Base):
    __tablename__ = "contact"

    id = Column(Integer, primary_key=True)

    address1 = Column(String(length=100))
    address2 = Column(String(length=100))
    address3 = Column(String(length=100))
    address4 = Column(String(length=100))
    address5 = Column(String(length=100))
    address6 = Column(String(length=100))
    alias = Column(String(length=255), default='')
    can_submit_commands = Column(Boolean)
    contact_name = Column(String(length=255))
    contactgroups = \
        relationship('Contactgroup', secondary=contact2contactgroup,
                     back_populates='contacts', lazy='select')
    custom_vars = \
        relationship('Customvar',
                     primaryjoin="and_(Contact.id==Customvar.id,"
                                 "Customvar.type=='contact')",
                     foreign_keys='[Customvar.id, Customvar.type]',
                     cascade='all, delete-orphan',
                     lazy='select',
                     overlaps='custom_vars')
    email = Column(String(length=255))
    file = relationship(File, lazy='select')
    hosts = \
        relationship('Host', secondary=host2contact,
                     back_populates='contacts', lazy='select')
    hostescalations = \
        relationship('Hostescalation', secondary=hostescalation2contact,
                     back_populates='contacts', lazy='select')
    minimum_value = Column(Integer)
    name = Column(String(length=255))
    pager = Column(String(length=100))
    register = Column(Boolean)
    retain_nonstatus_information = Column(Boolean)
    retain_status_information = Column(Boolean)
    services = \
        relationship('Service', secondary=service2contact,
                     back_populates='contacts', lazy='select')
    serviceescalations = \
        relationship('Serviceescalation', secondary=serviceescalation2contact,
                     back_populates='contacts', lazy='select')
    template = relationship('Contact', remote_side=id, lazy='select')

    _host_notif_per_fk = \
        Column('host_notification_period', Integer, ForeignKey(Timeperiod.id))
    _host_notif_cmd_fk = \
        Column('host_notification_cmds', Integer, ForeignKey(Command.id))

    host_notification_commands = \
        relationship(Command, foreign_keys=_host_notif_cmd_fk, lazy='select')
    host_notification_command_arguments = \
        Column('host_notification_cmds_args', Text)
    host_notifications_enabled = Column(Boolean)
    host_notification_options = Column(String(length=15))
    host_notification_timeperiod = \
        relationship(Timeperiod, foreign_keys=_host_notif_per_fk,
                     lazy='select')

    _serv_notif_per_fk = \
        Column('service_notification_period', Integer,
               ForeignKey(Timeperiod.id))
    _serv_notif_cmd_fk = \
        Column('service_notification_cmds', Integer, ForeignKey(Command.id))

    service_notification_command_arguments = \
        Column('service_notification_cmds_args', Text)
    service_notification_commands = \
        relationship(Command, foreign_keys=_serv_notif_cmd_fk, lazy='select')
    service_notifications_enabled = \
        Column(Boolean)
    service_notification_options = Column(String(length=15))
    service_notification_timeperiod = \
        relationship(Timeperiod, foreign_keys=_serv_notif_per_fk,
                     lazy='select')

    _file_fk = Column('file_id', Integer, ForeignKey(File.id))
    _template_fk = Column('template', Integer, ForeignKey(id))

    _file_id_syn = synonym('file')
    _host_notification_cmds_syn = synonym('host_notification_commands')
    _host_notification_cmds_args_syn = \
        synonym('host_notification_command_arguments')
    _host_notification_period_syn = synonym('host_notification_timeperiod')
    _service_notification_cmds_syn = synonym('service_notification_commands')
    _service_notification_cmds_args_syn = \
        synonym('service_notification_command_arguments')
    _service_notification_period_syn = \
        synonym('service_notification_timeperiod')

    @classmethod
    def exported_relationships(cls):
        return (
            cls.contactgroups,
            cls.custom_vars,
            cls.file,
            cls.host_notification_commands,
            cls.host_notification_timeperiod,
            cls.service_notification_commands,
            cls.service_notification_timeperiod,
            cls.template,
        )


class Contactgroup(Base):
    __tablename__ = "contactgroup"

    id = Column(Integer, primary_key=True)

    alias = Column(String(length=255), default='')
    contacts = \
        relationship('Contact', secondary=contact2contactgroup,
                     back_populates='contactgroups', lazy='select')
    file = relationship(File, lazy='joined')
    hosts = \
        relationship('Host', secondary=host2contactgroup,
                     back_populates='contactgroups', lazy='select')
    hostescalations = \
        relationship('Hostescalation', secondary=hostescalation2contactgroup,
                     back_populates='contactgroups', lazy='select')
    contactgroup_name = Column('contactgroup_name', String(length=255))
    register = Column(Boolean)
    services = \
        relationship('Service', secondary=service2contactgroup,
                     back_populates='contactgroups', lazy='select')
    serviceescalations = \
        relationship('Serviceescalation',
                     secondary=serviceescalation2contactgroup,
                     back_populates='contactgroups', lazy='select')
    subgroups = \
        relationship('Contactgroup',
                     secondary=contactgroup2member,
                     primaryjoin=id == contactgroup2member.c.contactgroup,
                     secondaryjoin=id == contactgroup2member.c.members,
                     back_populates='supergroups', lazy='select')
    supergroups = \
        relationship('Contactgroup',
                     secondary=contactgroup2member,
                     primaryjoin=id == contactgroup2member.c.members,
                     secondaryjoin=id == contactgroup2member.c.contactgroup,
                     back_populates='subgroups', lazy='select')

    _file_fk = Column('file_id', Integer, ForeignKey(File.id))

    _file_id_syn = synonym('file')
    _members_syn = synonym('contacts')
    _contactgroup_members_syn = synonym('subgroups')

    @classmethod
    def exported_relationships(cls):
        return (
            cls.file,
            cls.subgroups,
        )


class Customvar(Base):
    __tablename__ = "custom_vars"

    type = Column('obj_type', String(length=30), primary_key=True)
    id = Column('obj_id', Integer, primary_key=True)
    key = Column('variable', String(length=255), primary_key=True)
    value = Column(String(length=1024))

    _obj_type_syn = synonym('type')
    _id_syn = synonym('obj_id')
    _variable_syn = synonym('key')


class Host(Base):
    __tablename__ = "host"

    id = Column(Integer, primary_key=True)

    action_url = Column(String(length=2047))
    active_checks_enabled = Column(Boolean)
    address = Column(String(length=255), default='')
    alias = Column(String(length=255), default='')
    children = \
        relationship('Host', secondary=host2parent,
                     primaryjoin=id == host2parent.c.parents,
                     secondaryjoin=id == host2parent.c.host,
                     back_populates='parents', lazy='select')
    contactgroups = \
        relationship('Contactgroup', secondary=host2contactgroup,
                     back_populates='hosts', lazy='select')
    contacts = \
        relationship('Contact', secondary=host2contact,
                     back_populates='hosts', lazy='select')
    coords_2d = Column('2d_coords', String(length=20))
    custom_vars = \
        relationship('Customvar',
                     primaryjoin="and_(Host.id==Customvar.id,"
                                 "Customvar.type=='host')",
                     foreign_keys='[Customvar.id, Customvar.type]',
                     cascade='all, delete-orphan',
                     lazy='select',
                     overlaps='custom_vars')
    display_name = Column(String(length=255))
    file = relationship(File, lazy='select')
    first_notification_delay = Column(Integer)
    flap_detection_enabled = Column(Boolean)
    flap_detection_options = Column(String(length=18))
    freshness_threshold = Column(Float)
    high_flap_threshold = Column(Float)
    host_name = Column(String(length=255))
    hostgroups = \
        relationship('Hostgroup', secondary=host2hostgroup,
                     back_populates='hosts', lazy='select')
    hourly_value = Column(Integer)
    icon_image = Column(String(length=255))
    icon_image_alt = Column(String(length=255))
    low_flap_threshold = Column(Float)
    max_check_attempts = Column(SmallInteger)
    name = Column(String(length=255))
    notes = Column(Text)
    notes_url = Column(String(length=2047))
    obsess = Column(Boolean)
    parents = \
        relationship('Host', secondary=host2parent,
                     primaryjoin=id == host2parent.c.host,
                     secondaryjoin=id == host2parent.c.parents,
                     back_populates='children', lazy='select')
    passive_checks_enabled = Column(Boolean)
    process_perf_data = Column(Boolean)
    register = Column(Boolean)
    retain_nonstatus_information = Column(Boolean)
    retain_status_information = Column(Boolean)
    retry_interval = Column(SmallInteger)
    services = relationship("Service", back_populates="host", lazy="select")
    stalking_options = Column(String(length=15))
    statusmap_image = Column(String(length=255))
    template = relationship("Host", remote_side=id, lazy='select')

    _check_cmd_fk = Column('check_command', Integer, ForeignKey(Command.id))
    _check_per_fk = Column('check_period', Integer, ForeignKey(Timeperiod.id))

    check_command = relationship(Command, foreign_keys=_check_cmd_fk,
                                 lazy='select')
    check_command_arguments = Column('check_command_args', Text)
    check_freshness = Column(Boolean)
    check_interval = Column(SmallInteger)
    check_period = relationship(Timeperiod, foreign_keys=_check_per_fk,
                                lazy='select')

    _event_handler_fk = \
        Column('event_handler', Integer, ForeignKey(Command.id))

    event_handler = relationship(Command, foreign_keys=_event_handler_fk,
                                 lazy='select')
    event_handler_arguments = Column('event_handler_args', Text)
    event_handler_enabled = Column(Boolean)

    _notif_per_fk = \
        Column('notification_period', Integer, ForeignKey(Timeperiod.id))

    notification_timeperiod = \
        relationship(Timeperiod, foreign_keys=_notif_per_fk, lazy='select')
    notification_interval = Column(Integer)
    notification_options = Column(String(length=15))
    notifications_enabled = Column(Boolean)

    _file_fk = Column('file_id', Integer, ForeignKey(File.id))
    _template_fk = Column('template', Integer, ForeignKey(id))

    _2d_coords_syn = synonym('coords_2d')
    _check_command_args_syn = synonym('check_command_arguments')
    _contact_groups_syn = synonym('contactgroups')
    _event_handler_args_syn = synonym('event_handler_arguments')
    _file_id_syn = synonym('file')
    _notification_period_syn = synonym('notification_timeperiod')

    @classmethod
    def exported_relationships(cls):
        return (
            cls.check_command,
            cls.check_period,
            cls.contactgroups,
            cls.contacts,
            cls.custom_vars,
            cls.event_handler,
            cls.file,
            cls.hostgroups,
            cls.notification_timeperiod,
            cls.parents,
            cls.template,
        )


class Hostgroup(Base):
    __tablename__ = "hostgroup"

    id = Column(Integer, primary_key=True)

    action_url = Column(String(length=2047))
    alias = Column(String(length=255))
    custom_vars = \
        relationship('Customvar',
                     primaryjoin="and_(Hostgroup.id==Customvar.id,"
                                 "Customvar.type=='hostgroup')",
                     foreign_keys='[Customvar.id, Customvar.type]',
                     cascade='all, delete-orphan',
                     lazy='select',
                     overlaps='custom_vars')
    file = relationship(File, lazy='select')
    hosts = \
        relationship('Host', secondary=host2hostgroup,
                     back_populates='hostgroups', lazy='select')
    hostgroup_name = Column(String(length=255))
    notes = Column(Text)
    notes_url = Column(String(length=2047))
    register = Column(Boolean)
    subgroups = \
        relationship('Hostgroup', secondary=hostgroup2member,
                     primaryjoin=id == hostgroup2member.c.hostgroup,
                     secondaryjoin=id == hostgroup2member.c.members,
                     back_populates='supergroups', lazy='select')
    supergroups = \
        relationship('Hostgroup', secondary=hostgroup2member,
                     primaryjoin=id == hostgroup2member.c.members,
                     secondaryjoin=id == hostgroup2member.c.hostgroup,
                     back_populates='subgroups', lazy='select')

    _file_fk = Column('file_id', Integer, ForeignKey(File.id))

    _file_id_syn = synonym('file')
    _hostgroup_members_syn = synonym('subgroups')
    _members_syn = synonym('hosts')

    @classmethod
    def exported_relationships(cls):
        return (
            cls.custom_vars,
            cls.file,
            cls.subgroups,
        )


class Hostdependency(Base):
    __tablename__ = "hostdependency"

    id = Column(Integer, primary_key=True)

    dependency_period = relationship(Timeperiod, lazy='select')
    execution_failure_criteria = Column(String(length=15))
    file = relationship(File, lazy='select')
    inherits_parent = Column(Boolean)
    notification_failure_criteria = Column(String(length=15))
    register = Column(Boolean)

    _depen_host_fk = \
        Column('dependent_host_name', Integer, ForeignKey(Host.id),
               primary_key=True, default=0)
    _depen_hostgroup_fk = \
        Column('dependent_hostgroup_name', Integer, ForeignKey(Hostgroup.id),
               primary_key=True, default=0)

    dependent_host = relationship(Host, foreign_keys=_depen_host_fk,
                                  lazy='select')
    dependent_hostgroup = \
        relationship(Hostgroup, foreign_keys=_depen_hostgroup_fk,
                     lazy='select')

    _host_fk = Column('host_name', Integer, ForeignKey(Host.id),
                      primary_key=True, default=0)
    _hostgroup_fk = Column('hostgroup_name', Integer, ForeignKey(Hostgroup.id),
                           primary_key=True, default=0)

    host = relationship(Host, foreign_keys=_host_fk, lazy='select')
    hostgroup = relationship(Hostgroup, foreign_keys=_hostgroup_fk,
                             lazy='select')

    _depen_per_fk = \
        Column('dependency_period', Integer, ForeignKey(Timeperiod.id))
    _file_fk = Column('file_id', Integer, ForeignKey(File.id))

    _dependent_host_name_syn = synonym('dependent_host')
    _dependent_hostgroup_name_syn = synonym('dependent_hostgroup')
    _file_id_syn = synonym('file')
    _host_name_syn = synonym('host')
    _hostgroup_name_syn = synonym('hostgroup')

    @classmethod
    def exported_relationships(cls):
        return (
            cls.dependency_period,
            cls.dependent_host,
            cls.dependent_hostgroup,
            cls.file,
            cls.host,
            cls.hostgroup,
        )


class Hostescalation(Base):
    __tablename__ = "hostescalation"

    id = Column(Integer, primary_key=True)

    contacts = \
        relationship('Contact', secondary=hostescalation2contact,
                     back_populates='hostescalations', lazy='select')
    contactgroups = \
        relationship('Contactgroup', secondary=hostescalation2contactgroup,
                     back_populates='hostescalations', lazy='select')
    escalation_options = Column(String(length=15))
    escalation_timeperiod = relationship(Timeperiod, lazy='select')
    file = relationship(File, lazy='select')
    first_notification = Column(Integer)
    host = relationship(Host, lazy='select')
    hostgroup = relationship(Hostgroup, lazy='select')
    last_notification = Column(Integer)
    notification_interval = Column(Integer)
    register = Column(Boolean)

    _escal_per_fk = \
        Column('escalation_period', Integer, ForeignKey(Timeperiod.id))
    _file_fk = Column('file_id', Integer, ForeignKey(File.id))
    _host_fk = Column('host_name', Integer, ForeignKey(Host.id),
                      primary_key=True, default=0)
    _hostgroup_fk = Column('hostgroup_name', Integer, ForeignKey(Hostgroup.id),
                           primary_key=True, default=0)

    _file_id_syn = synonym('file')
    _escalation_period_syn = synonym('escalation_timeperiod')
    _host_name_syn = synonym('host')
    _hostgroup_name_syn = synonym('hostgroup')
    _contact_groups_syn = synonym('contactgroups')

    @classmethod
    def exported_relationships(cls):
        return (
            cls.contactgroups,
            cls.contacts,
            cls.escalation_timeperiod,
            cls.file,
            cls.host,
            cls.hostgroup,
        )


class Service(Base):
    __tablename__ = "service"

    id = Column(Integer, primary_key=True)

    action_url = Column(String(length=2047))
    active_checks_enabled = Column(Boolean)
    contacts = \
        relationship('Contact', secondary=service2contact,
                     back_populates='services', lazy='select')
    contactgroups = \
        relationship('Contactgroup', secondary=service2contactgroup,
                     back_populates='services', lazy='select')
    custom_vars = \
        relationship('Customvar',
                     primaryjoin="and_(Service.id==Customvar.id,"
                                 "Customvar.type=='service')",
                     foreign_keys='[Customvar.id, Customvar.type]',
                     cascade='all, delete-orphan',
                     lazy='select',
                     overlaps='custom_vars')
    description = Column('service_description', String(length=255))
    display_name = Column(String(length=255))
    file = relationship(File, lazy='select')
    first_notification_delay = Column(Integer)
    flap_detection_enabled = Column(Boolean)
    flap_detection_options = Column(String(length=18))
    freshness_threshold = Column(Integer)
    high_flap_threshold = Column(Float)
    host = relationship(Host, lazy='select', back_populates="services")
    hostgroup = relationship(Hostgroup, lazy='select')
    hourly_value = Column(Integer)
    icon_image = Column(String(length=255))
    icon_image_alt = Column(String(length=255))
    is_volatile = Column(Boolean)
    low_flap_threshold = Column(Float)
    max_check_attempts = Column(SmallInteger)
    name = Column(String(length=255))
    notes = Column(Text)
    notes_url = Column(String(length=2047))
    obsess = Column(Boolean)
    parallelize_check = Column(Boolean)
    passive_checks_enabled = Column(Boolean)
    process_perf_data = Column(Boolean)
    register = Column(Boolean)
    retain_nonstatus_information = Column(Boolean)
    retain_status_information = Column(Boolean)
    retry_interval = Column(SmallInteger)
    stalking_options = Column(String(length=15))
    servicegroups = \
        relationship('Servicegroup', secondary=service2servicegroup,
                     back_populates='services', lazy='select')
    template = relationship("Service", remote_side=id, lazy='select')

    _check_cmd_fk = Column('check_command', Integer, ForeignKey(Command.id))
    _check_per_fk = Column('check_period', Integer, ForeignKey(Timeperiod.id))

    check_command = relationship(Command, foreign_keys=_check_cmd_fk,
                                 lazy='select')
    check_command_arguments = Column('check_command_args', Text)
    check_freshness = Column(Boolean)
    check_interval = Column(SmallInteger)
    check_period = relationship(Timeperiod, foreign_keys=_check_per_fk,
                                lazy='select')

    _event_handler_fk = \
        Column('event_handler', Integer, ForeignKey(Command.id))
    event_handler = relationship(Command, foreign_keys=_event_handler_fk,
                                 lazy='select')
    event_handler_arguments = Column('event_handler_args', Text)
    event_handler_enabled = Column(Boolean)

    _notif_per_fk = \
        Column('notification_period', Integer, ForeignKey(Timeperiod.id))
    notifications_enabled = Column(Boolean)
    notification_interval = Column(Integer)
    notification_options = Column(String(length=15))
    notification_timeperiod = \
        relationship(Timeperiod, foreign_keys=_notif_per_fk,
                     lazy='select')

    _file_fk = Column('file_id', Integer, ForeignKey(File.id))
    _host_fk = Column('host_name', Integer, ForeignKey(Host.id))
    _hostgroup_fk = Column('hostgroup_name', Integer, ForeignKey(Hostgroup.id))
    _template_fk = Column('template', Integer, ForeignKey(id))

    _file_id_syn = synonym('file')
    _host_name_syn = synonym('host')
    _notification_period_syn = synonym('notification_timeperiod')
    _event_handler_args_syn = synonym('event_handler_arguments')
    _contact_groups_syn = synonym('contactgroups')
    _check_command_args_syn = synonym('check_command_arguments')
    _service_description_syn = synonym('description')
    _hostgroup_name_syn = synonym('hostgroup')

    @classmethod
    def exported_relationships(cls):
        return (
            cls.check_command,
            cls.check_period,
            cls.contactgroups,
            cls.contacts,
            cls.custom_vars,
            cls.event_handler,
            cls.file,
            cls.host,
            cls.hostgroup,
            cls.notification_timeperiod,
            cls.servicegroups,
            cls.template,
        )


class Servicedependency(Base):
    __tablename__ = "servicedependency"

    id = Column(Integer, primary_key=True)

    dependency_period = relationship(Timeperiod, lazy='select')
    execution_failure_criteria = Column(String(length=15))
    file = relationship(File, lazy='select')
    inherits_parent = Column(Boolean)
    notification_failure_criteria = Column(String(length=15))
    register = Column(Boolean)

    _depen_serv_fk = \
        Column('dependent_service', Integer, ForeignKey(Service.id))

    dependent_service = relationship(Service, foreign_keys=_depen_serv_fk,
                                     lazy='select')

    _serv_fk = Column('service', Integer, ForeignKey(Service.id))

    service = relationship(Service, foreign_keys=_serv_fk, lazy='select')

    _depen_per_fk = \
        Column('dependency_period', Integer, ForeignKey(Timeperiod.id))
    _file_fk = Column('file_id', Integer, ForeignKey(File.id))

    _file_id_syn = synonym('file')

    @classmethod
    def exported_relationships(cls):
        return (
            cls.dependency_period,
            cls.dependent_service,
            cls.file,
            cls.service,
        )


class Serviceescalation(Base):
    __tablename__ = "serviceescalation"

    id = Column(Integer, primary_key=True)

    contacts = \
        relationship('Contact', secondary=serviceescalation2contact,
                     back_populates='serviceescalations', lazy='select')
    contactgroups = \
        relationship('Contactgroup', secondary=serviceescalation2contactgroup,
                     back_populates='serviceescalations', lazy='select')
    escalation_options = Column(String(length=15))
    escalation_period = relationship(Timeperiod, lazy='select')
    file = relationship(File, lazy='select')
    first_notification = Column(Integer)
    last_notification = Column(Integer)
    notification_interval = Column(Integer)
    register = Column(Boolean)
    service = relationship(Service, lazy='select')

    _escal_per_fk = \
        Column('escalation_period', Integer, ForeignKey(Timeperiod.id))
    _file_fk = Column('file_id', Integer, ForeignKey(File.id))
    _serv_fk = Column('service', Integer, ForeignKey(Service.id))

    _contact_groups_syn = synonym('contactgroups')
    _file_id_syn = synonym('file')

    @classmethod
    def exported_relationships(cls):
        return (
            cls.contactgroups,
            cls.contacts,
            cls.escalation_period,
            cls.file,
            cls.service,
        )


class Servicegroup(Base):
    __tablename__ = "servicegroup"

    id = Column(Integer, primary_key=True)

    action_url = Column(String(length=2047))
    alias = Column(String(length=255), default='')
    custom_vars = \
        relationship('Customvar',
                     primaryjoin="and_(Servicegroup.id==Customvar.id,"
                                 "Customvar.type=='servicegroup')",
                     foreign_keys='[Customvar.id, Customvar.type]',
                     cascade='all, delete-orphan',
                     lazy='select',
                     overlaps='custom_vars')
    file = relationship(File, lazy='select')
    name = Column('servicegroup_name', String(length=255))
    notes = Column(Text)
    notes_url = Column(String(length=2047))
    services = \
        relationship('Service', secondary=service2servicegroup,
                     back_populates='servicegroups', lazy='select')
    register = Column(Boolean)
    subgroups = \
        relationship('Servicegroup', secondary=servicegroup2member,
                     primaryjoin=id == servicegroup2member.c.servicegroup,
                     secondaryjoin=id == servicegroup2member.c.members,
                     back_populates='supergroups', lazy='select')
    supergroups = \
        relationship('Servicegroup', secondary=servicegroup2member,
                     primaryjoin=id == servicegroup2member.c.members,
                     secondaryjoin=id == servicegroup2member.c.servicegroup,
                     back_populates='subgroups', lazy='select')
    _file_fk = Column('file_id', Integer, ForeignKey(File.id))

    _file_id_syn = synonym('file')
    _members_syn = synonym('services')
    _servicegroup_members_syn = synonym('subgroups')
    _servicegroup_name_syn = synonym('name')

    @classmethod
    def exported_relationships(cls):
        return (cls.custom_vars, cls.file, cls.subgroups)


class Objlock(Base):
    __tablename__ = "objlocks"

    id = Column(Integer, primary_key=True, autoincrement=True)
    object_type = Column(String(length=100))
    object_id = Column(Integer)
    username = Column(String(length=255))
    time = Column(Epoch)
