from nachos.nacoma.api.sqlalchemy.models import Timeperiod
from nachos.nacoma.mappers.sqlalchemy.mapper import Mapper
from nachos.nacoma.mappers.base.changelog import BaseChangelogMapper

from oslo_config import cfg
from phpserialize import loads as php_loads


class ChangelogMapper(Mapper, BaseChangelogMapper):

    def __init__(self, sqlalchemy_obj):
        Mapper.__init__(self, sqlalchemy_obj)
        if self.oldstate:
            self._decoded_old_state = php_loads(
                self.oldstate,
                charset=cfg.CONF.nacoma.blob_encoding,
                decode_strings=True,
            )
        else:
            self._decoded_old_state = {}
        if self.newstate:
            self._decoded_new_state = php_loads(
                self.newstate,
                charset=cfg.CONF.nacoma.blob_encoding,
                decode_strings=True,
            )
        else:
            self._decoded_new_state = {}
        self._is_deleted = False
        self._is_new = False

    @property
    def object_type(self):
        return self.mapattr("object_type")

    @property
    def object_id(self):
        return self.mapattr("object_id")

    @property
    def object_identifier(self):
        return (self.object_type, self.object_id)

    @property
    def oldname(self):
        return self.mapattr("oldname")

    @property
    def newname(self):
        return self.mapattr("newname")

    @property
    def oldstate(self):
        return self.mapattr("oldstate")

    @property
    def newstate(self):
        return self.mapattr("newstate")

    @property
    def username(self):
        return self.mapattr("username")

    @property
    def time(self):
        return self.mapattr("time")

    @property
    def is_new(self):
        return bool(
            self._is_new
            or (not self.oldname and self.newname)
            or (
                not self._decoded_old_state
                and self._decoded_new_state
                and not self.is_custom_var
            )
        )

    @property
    def is_deleted(self):
        return bool(
            self._is_deleted
            or (self.oldname and not self.newname)
            or (
                self._decoded_old_state
                and not self._decoded_new_state
                and not self.is_custom_var
            )
        )

    @property
    def is_renamed(self):
        # These objects don't have names
        if self.object_type in [
            "hostdependency",
            "servicedependency",
            "hostescalation",
            "serviceescalation",
        ]:
            return False
        return bool(
            self.newname
            and self.oldname
            and self._decoded_old_state
            and self._decoded_new_state
            and self.oldname != self.newname
        )

    @property
    def old_file_id(self):
        file_id = self._decoded_old_state.get('file_id')
        if file_id:
            return int(file_id)
        return None

    @property
    def new_file_id(self):
        file_id = self._decoded_new_state.get('file_id')
        if file_id:
            return int(file_id)
        return None

    @property
    def is_moved(self):
        return (
            (self.old_file_id is not None)
            and (self.new_file_id is not None)
            and (self.new_file_id != self.old_file_id)
        )

    def get_from_oldstate(self, attr):
        return self._decoded_old_state.get(attr)

    def get_from_newstate(self, key):
        return self._decoded_new_state.get(key)

    def get_oldstate_members(self, key, type=int):
        return self._get_state_members(self.get_from_oldstate(key), type)

    def get_newstate_members(self, key, type=int):
        return self._get_state_members(self.get_from_newstate(key), type)

    @staticmethod
    def _get_state_members(members, type=int):
        if members:
            return [type(member) for member in members.values()]
        return None

    def get_all_state_members(self, key, type=int):
        old_members = self.get_oldstate_members(key, type) or []
        new_members = self.get_newstate_members(key, type) or []
        all_members = set(old_members)
        all_members.update(new_members)
        return all_members

    @property
    def is_custom_var(self):
        old_state = self._decoded_old_state
        new_state = self._decoded_new_state
        if not old_state and not new_state:
            return False
        state = old_state if old_state else new_state
        for key in state.keys():
            if key.startswith('_') and key.isupper():
                return True
            if (self.object_type == "timeperiod"
                    and not hasattr(Timeperiod, key)
                    and not hasattr(Timeperiod, "_{}_syn".format(key))):
                return True
        return False

    def merge_change(self, other):
        self._is_new = self.is_new or other.is_new
        self._is_deleted = self.is_deleted or other.is_deleted
        if not self.is_deleted:
            self._decoded_new_state.update(other._decoded_new_state)
        if other.is_renamed:
            self.sqlalchemy_obj.newname = other.newname
        old_state = other._decoded_old_state
        old_state.update(self._decoded_old_state)
        self._decoded_old_state = old_state
